from pykechain.exceptions import _DeprecationMixin


class SelectListProperty2(_DeprecationMixin):
    """A virtual object representing a KE-chain single-select list property."""

    pass


class MultiSelectListProperty2(_DeprecationMixin):
    """A virtual object representing a KE-chain multi-select list property."""

    pass



import warnings
from datetime import datetime
from typing import Text, Optional, List, Iterable, Any, Callable, Dict

from pykechain.enums import Enum
from pykechain.exceptions import IllegalArgumentError
from pykechain.utils import is_uuid, is_url, Empty, empty

iter_types = (list, tuple, set)


def check_type(value: Optional[Any], cls: Any, key: Text) -> Optional[Any]:
    """Validate any input to be an instance a specific class.

    :param value: value to check against
    :param cls: check to see if the value is of type class
    :param key: key
    """
    if value is not None and value is not empty:
        if not isinstance(value, cls):
            if isinstance(cls, iter_types):
                types = ', '.join(c.__name__ for c in cls)
            else:
                types = cls.__name__
            raise IllegalArgumentError('`{}` should be of type {}, "{}" ({}) is not.'.format(
                key, types, value, type(value)))
    return value


def check_uuid(uuid: Optional[Text], key: Optional[Text] = 'pk') -> Optional[Text]:
    """Validate the UUID input to be a correct UUID."""
    if uuid is not None and uuid is not empty:
        if not is_uuid(uuid):
            raise IllegalArgumentError('`{}` must be a valid UUID, "{}" ({}) is not.'.format(key, uuid, type(uuid)))
    return uuid


def check_text(text: Optional[Text], key: Text) -> Optional[Text]:
    """Validate text input to be a string."""
    if text is not None and text is not empty:
        if not isinstance(text, str):
            raise IllegalArgumentError('`{}` should be a string, "{}" ({}) is not.'.format(key, text, type(text)))
    return text


def check_url(url: Optional[Text], key: Text = 'url') -> Optional[Text]:
    """Validate text input to be a valid format URL."""
    url = check_text(text=url, key=key)
    if url is not None and url is not empty:
        if not is_url(url):
            raise IllegalArgumentError('`{}` should be a valid URL, "{}" ({}) is not.'.format(key, url, type(url)))
    return url


def check_list_of_text(list_of_text: Optional[Iterable[Text]], key: Text, unique: bool = False) -> Optional[List[Text]]:
    """Validate iterable input to be a list/tuple/set of strings."""
    if list_of_text is not None and list_of_text is not empty:
        if not isinstance(list_of_text, iter_types) or not all(isinstance(t, Text) for t in list_of_text):
            raise IllegalArgumentError(
                '`{}` should be a list, tuple or set of strings, "{}" ({}) is not.'.format(key, list_of_text,
                                                                                           type(list_of_text)))
        if unique:
            list_of_text = [t for i, t in enumerate(list_of_text) if list_of_text.index(t) == i]
    return list_of_text


def check_list_of_dicts(list_of_dicts: Optional[Iterable[Dict]],
                        key: Text,
                        fields: Optional[List[Text]] = None,
                        ) -> Optional[List[Dict]]:
    """Validate iterable input to be a list/tuple/set of dicts, optionally checking for required field names."""
    if list_of_dicts is not None and list_of_dicts is not empty:
        if not isinstance(list_of_dicts, iter_types) or not all(isinstance(d, dict) for d in list_of_dicts):
            raise IllegalArgumentError(
                '`{}` should be a list, tuple or set of dicts, "{}" ({}) is not.'.format(key, list_of_dicts,
                                                                                         type(list_of_dicts))
            )
        if fields:
            assert isinstance(fields, list) and all(isinstance(f, str) for f in fields), \
                '`fields` must be a list of strings.'
            missing_fields = set()
            for dictionary in list_of_dicts:
                for field in fields:
                    if field not in dictionary:
                        missing_fields.add(field)
            if missing_fields:
                raise IllegalArgumentError('Not every dict contains the required fields: "{}"\n'
                                           'Missing fields: "{}"'.format('", "'.join(fields),
                                                                         '", "'.join(list(missing_fields))))
    return list_of_dicts


def check_enum(value: Optional[Any], enum: type(Enum), key: Text) -> Optional[Any]:
    """Validate input to be an option from an enum class."""
    if value is not None and value is not empty:
        if value not in enum.values():
            raise IllegalArgumentError('`{}` must be an option from enum {}, "{}" ({}) is not.\nChoose from: {}'.format(
                key, enum.__class__.__name__, value, type(value), enum.values()))
    return value


def check_datetime(dt: Optional[datetime], key: Text) -> Optional[Text]:
    """Validate a datetime value to be a datetime and be timezone aware."""
    if dt is not None and dt is not empty:
        if isinstance(dt, datetime):
            if not dt.tzinfo:
                warnings.warn("`{}` '{}' is naive and not timezone aware, use pytz.timezone info. "
                              "Date will be interpreted as UTC time.".format(key, dt.isoformat(sep=' ')))
            dt = dt.isoformat(sep='T')
        else:
            raise IllegalArgumentError(
                '`{}` should be a datetime.datetime() object, "{}" ({}) is not.'.format(key, dt, type(dt)))
    return dt


def check_base(obj: Optional[Any],
               cls: Optional[type(object)] = None,
               key: Optional[Text] = 'object',
               method: Optional[Callable] = None,
               ) -> Optional[Text]:
    """Validate whether the object provided as input is a Base (or subclass) instance and return its ID."""
    if obj is not None and obj is not empty:

        if cls is None:
            from pykechain.models import Base
            cls = Base

        if isinstance(obj, cls):
            obj = obj.id
        elif isinstance(obj, str) and is_uuid(obj):
            pass
        elif isinstance(obj, str) and method:
            obj = method(obj).id
        else:
            raise IllegalArgumentError(
                '`{}` must be an ID, UUID or `{}` object, "{}" ({}) is not.'.format(key, cls.__name__, obj, type(obj)))
    return obj


def check_user(obj: Optional[Any],
               cls: Optional[type(object)] = None,
               key: Optional[Text] = 'user',
               method: Optional[Callable] = None,
               ) -> Optional[int]:
    """Provide same functionality as check_base(), although users dont use UUID but integers."""
    if obj is not None and obj is not empty:

        if cls is None:
            from pykechain.models import User
            cls = User

        if isinstance(obj, cls):
            obj = obj.id
        elif isinstance(obj, (int, str)):
            try:
                obj = int(obj)
            except ValueError:
                raise IllegalArgumentError(
                    '`{}` must be an ID or `{}` object, "{}" ({}) is not.'.format(key, cls.__name__, obj, type(obj)))
        elif method:
            obj = method(obj).id
        else:
            raise IllegalArgumentError(
                '`{}` must be an ID or `{}` object, "{}" ({}) is not.'.format(key, cls.__name__, obj, type(obj)))
    return obj


def check_list_of_base(
    objects: Optional[List[Any]],
    cls: Optional[type(object)] = None,
    key: Optional[Text] = 'objects',
    method: Optional[Callable] = None,
) -> Optional[List[Text]]:
    """Validate the iterable of objects provided as input are Base (or subclass) instances and return a list of IDs.

    :param objects: list of objects to check
    :param cls: (optional) class to check objects against
    :param key: (optional) key to check
    :param method: (optional) method used to check with
    :return: list of UUID's in Text.
    """
    ids = None
    if objects is not None and objects is not empty:
        check_type(objects, iter_types, key=key)

        if cls is None:
            from pykechain.models import Base
            cls = Base

        ids = []
        not_recognized = []
        for obj in objects:
            try:
                ids.append(check_base(obj=obj, cls=cls, method=method))
            except IllegalArgumentError:
                not_recognized.append(obj)

        if not_recognized:
            raise IllegalArgumentError(
                'All `{}` must be IDs, UUIDs or `{}` objects, "{}" is/are not.'.format(
                    key, cls.__name__, '", "'.join([str(n) for n in not_recognized]))
            )
    return ids


def check_empty(value: Optional[Any]):
    """Validate whether the value provided is of class `Empty`."""
    if isinstance(value, Empty):
        return True
    return False



import os

from envparse import env

from pykechain.client import Client
from pykechain.enums import KechainEnv as kecenv, ScopeStatus
from pykechain.exceptions import ClientError


def get_project(url=None, username=None, password=None, token=None, scope=None, scope_id=None,
                env_filename=None, status=ScopeStatus.ACTIVE, check_certificates=None):
    """
    Retrieve and return the KE-chain project to be used throughout an app.

    This helper is made to bootstrap a pykechain enabled python script or an jupyter notebook with the correct
    project (technically this is a `pykechain.models.Scope` model).

    When no parameters are passed in this function, it will try to retrieve `url`, `token`, `scope` (or `scope_id`)
    from the environment variables or a neatly placed '.env' file.

    when the environment variable KECHAIN_FORCE_ENV_USE is set to true, (or ok, on, 1, yes) then the use of
    environmentvariables for the retrieval of the scope are enforced. The following environment variables can be set::

        KECHAIN_URL           - full url of KE-chain where to connect to eg: 'https://<some>.ke-chain.com'
        KECHAIN_TOKEN         - authentication token for the KE-chain user provided from KE-chain user account control
        KECHAIN_USERNAME      - the username for the credentials
        KECHAIN_PASSWORD      - the password for the credentials
        KECHAIN_SCOPE         - the name of the project / scope. Should be unique, otherwise use scope_id
        KECHAIN_SCOPE_ID      - the UUID of the project / scope.
        KECHAIN_FORCE_ENV_USE - set to 'true', '1', 'ok', or 'yes' to always use the environment variables.
        KECHAIN_SCOPE_STATUS  - the status of the Scope to retrieve, defaults to None to retrieve all scopes
        KECHAIN_CHECK_CERTIFICATES - if the certificates of the URL should be checked. default to True if not present.

    .. versionadded:: 1.12

    :param url: (optional) url of KE-chain
    :type url: basestring or None
    :param username: (optional) username for authentication (together with password, if not token)
    :type username: basestring or None
    :param password: (optional) password for username/password authentication (together with username, if not token)
    :type password: basestring or None
    :param token: (optional) token for authentication (if not username/password)
    :type token: basestring or None
    :param scope: (optional) name of the scope to retrieve from KE-chain.
    :type scope: basestring or None
    :param scope_id: (optional) UUID of the scope to retrieve and return from KE-chain
    :type scope_id: basestring or None
    :param env_filename: (optional) name of the environment filename to bootstrap the Client
    :type env_filename: basestring or None
    :param status: (optional) status of the scope to retrieve, defaults to :attr:`enums.Scopestatus.ACTIVE`
    :type status: basestring or None
    :param check_certificates: (optional) if to check TLS/SSL Certificates. If nothing provided (None) it
                               ensures it does check certficates.
    :type check_certificates: bool or None
    :return: pykechain.models.Scope
    :raises NotFoundError: If the scope could not be found
    :raises ClientError: If the client connection to KE-chain was unsuccessful
    :raises APIError: If other Errors occur to retrieve the scope

    Example
    -------
    An example with parameters provided

    >>> from pykechain import get_project
    >>> project = get_project(url='http://localhost:8000',
    ...     username='foo', password='bar', scope='1st!')
    >>> print(project.name)
    1st

    An example with a .env file on disk::

        # This is an .env file on disk.
        KECHAIN_TOKEN=bd9377793f7e74a29dbb11fce969
        KECHAIN_URL=http://localhost:8080
        KECHAIN_CHECK_CERTIFICATES=False
        KECHAIN_SCOPE_ID=c9f0-228e-4d3a-9dc0-ec5a75d7

    >>> project = get_project(env_filename='/path/to/.env')
    >>> project.id
    c9f0-228e-4d3a-9dc0-ec5a75d7

    An example for get_project that will extract all from the environment variables

    >>> env_vars = os.environ
    >>> env_vars.get('KECHAIN_TOKEN')
    bd9377793f7e74a29dbb11fce969
    >>> env_vars.get('KECHAIN_URL')
    http://localhost:8080
    >>> env_vars.get('KECHAIN_SCOPE')
    Bike Project
    >>> project = get_project()
    >>> project.name
    Bike Project

    """
    if env.bool(kecenv.KECHAIN_FORCE_ENV_USE, default=False):
        if not os.getenv(kecenv.KECHAIN_URL):
            raise ClientError(
                "Error: KECHAIN_URL should be provided as environment variable (use of env vars is enforced)")
        if not (os.getenv(kecenv.KECHAIN_TOKEN) or  # noqa: W504
                (os.getenv(kecenv.KECHAIN_PASSWORD) and os.getenv(kecenv.KECHAIN_PASSWORD))):
            raise ClientError("Error: KECHAIN_TOKEN or KECHAIN_USERNAME and KECHAIN_PASSWORD should be provided as "
                              "environment variable(s) (use of env vars is enforced)")
        if not (os.getenv(kecenv.KECHAIN_SCOPE) or os.getenv(kecenv.KECHAIN_SCOPE_ID)):
            raise ClientError("Error: KECHAIN_SCOPE or KECHAIN_SCOPE_ID should be provided as environment variable "
                              "(use of env vars is enforced)")

    if env.bool(kecenv.KECHAIN_FORCE_ENV_USE, default=False) or \
            not any((url, username, password, token, scope, scope_id)):
        check_certificates = env.bool(kecenv.KECHAIN_CHECK_CERTIFICATES, default=True)
        client = Client.from_env(env_filename=env_filename, check_certificates=check_certificates)
        scope_id = env(kecenv.KECHAIN_SCOPE_ID, default=None)
        scope = env(kecenv.KECHAIN_SCOPE, default=None)
        status = env(kecenv.KECHAIN_SCOPE_STATUS, default=None)
    elif (url and ((username and password) or (token)) and (scope or scope_id)) and \
            not env.bool(kecenv.KECHAIN_FORCE_ENV_USE, default=False):
        if check_certificates is None:
            check_certificates = env.bool(kecenv.KECHAIN_CHECK_CERTIFICATES, default=True)
        client = Client(url=url, check_certificates=check_certificates)
        client.login(username=username, password=password, token=token)
    else:
        raise ClientError("Error: insufficient arguments to connect to KE-chain. "
                          "See documentation of `pykechain.get_project()`")

    if scope_id:
        return client.scope(pk=scope_id, status=status)
    else:
        return client.scope(name=scope, status=status)



import os
from typing import List
from unittest import TestCase

from pykechain.enums import (WidgetTypes, ShowColumnTypes, FilterType, ProgressBarColors,
                             Category, LinkTargets, KEChainPages, WidgetTitleValue, Alignment, ActivityType,
                             CardWidgetLinkValue, CardWidgetLinkTarget, ImageFitValue, PropertyType, Classification,
                             Multiplicity)
from pykechain.models.widgets.enums import (
    DashboardWidgetShowTasks,
    DashboardWidgetShowScopes,
)
from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models import Activity, Part
from pykechain.models.widgets import (
    UndefinedWidget, HtmlWidget, PropertygridWidget, AttachmentviewerWidget, SupergridWidget, FilteredgridWidget,
    TasknavigationbarWidget, SignatureWidget, ServiceWidget, NotebookWidget, MulticolumnWidget, CardWidget,
    MetapanelWidget, ScopeWidget)
from pykechain.models.widgets.helpers import _set_title
from pykechain.models.widgets.widget import Widget
from pykechain.models.widgets.widget_models import ServicecardWidget, DashboardWidget
from pykechain.models.widgets.widgets_manager import WidgetsManager
from pykechain.utils import slugify_ref, temp_chdir, find
from tests.classes import TestBetamax


class TestSetTitle(TestCase):

    def test_interface(self):
        title_in = 'title'
        meta_in = dict()
        output = _set_title(meta_in, title_in)

        self.assertIsInstance(output, tuple)
        self.assertTrue(len(output) == 2)

        meta, title = output

        self.assertIsInstance(meta, dict)
        self.assertIsInstance(title, str)
        self.assertIs(meta_in, meta, msg='Meta must be updated in-place!')
        self.assertIn('showTitleValue', meta)
        self.assertIn('customTitle', meta)

    def test_title(self):
        """Test parser when changing the `title` input."""
        title_in = 'title'
        meta, title = _set_title(dict(), title=title_in)
        self.assertEqual(title_in, title)
        self.assertEqual(title_in, meta['customTitle'])
        self.assertEqual(WidgetTitleValue.CUSTOM_TITLE, meta['showTitleValue'])

        title_in = False
        meta, title = _set_title(dict(), title=title_in)
        self.assertIsNone(title)
        self.assertIsNone(meta['customTitle'])
        self.assertEqual(WidgetTitleValue.DEFAULT, meta['showTitleValue'])

        title_in = None
        meta, title = _set_title(dict(), title=title_in)
        self.assertIsNone(title)
        self.assertIsNone(meta['customTitle'])
        self.assertEqual(WidgetTitleValue.NO_TITLE, meta['showTitleValue'])

    def test_show_title_value(self):
        """Test parser when changing the `show_title_value` input."""
        title_in = 'title'

        meta, title = _set_title(
            dict(), title=title_in, show_title_value=WidgetTitleValue.NO_TITLE)

        self.assertEqual(title_in, title)
        self.assertEqual(title_in, meta['customTitle'])
        self.assertEqual(WidgetTitleValue.NO_TITLE, meta['showTitleValue'])

        meta, title = _set_title(
            dict(), title=title_in, show_title_value=WidgetTitleValue.CUSTOM_TITLE)

        self.assertEqual(title_in, title)
        self.assertEqual(title_in, meta['customTitle'])
        self.assertEqual(WidgetTitleValue.CUSTOM_TITLE, meta['showTitleValue'])

        meta, title = _set_title(
            dict(), title=title_in, show_title_value=WidgetTitleValue.DEFAULT)

        self.assertIsNone(title)
        self.assertIsNone(meta['customTitle'])
        self.assertEqual(WidgetTitleValue.DEFAULT, meta['showTitleValue'])

        with self.assertRaises(IllegalArgumentError, msg='Unrecognized show_title_value must be caught!'):
            # noinspection PyTypeChecker
            _set_title(dict(), 'title', show_title_value='Maybe')


class TestWidgets(TestBetamax):

    def setUp(self):
        super().setUp()

        self.activity = self.project.activity('Specify wheel diameter')
        self.new_widget = None

    def tearDown(self):
        if self.new_widget:
            self.new_widget.delete()
        super().tearDown()

    def test_retrieve_widgets_in_activity(self):
        widget_set = self.activity.widgets()

        self.assertIsInstance(widget_set, WidgetsManager)

        for w in widget_set:
            with self.subTest(msg=w):
                self.assertIsInstance(w, Widget)
                self.assertEqual(self.activity, w.activity())

    def test_create_widget_in_activity(self):
        self.new_widget = self.client.create_widget(
            title="Test_widget",
            activity=self.activity,
            widget_type=WidgetTypes.UNDEFINED,
            meta={},
            order=100
        )
        self.assertIsInstance(self.new_widget, UndefinedWidget)

    def test_widget_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'created_at', 'updated_at', 'ref',
                      'widget_type', 'title', 'meta', 'order', '_activity_id', '_parent_id',
                      'has_subwidgets', '_scope_id', 'progress']
        first_widget = self.activity.widgets()[0]

        self.assertIsInstance(first_widget, Widget)
        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(hasattr(first_widget, attribute))

    def test_widget_meta_attribute_is_not_None(self):
        first_widget = self.activity.widgets()[0]

        self.assertIsInstance(first_widget, Widget)
        self.assertIsNotNone(first_widget.meta)

    def test_associated_parts(self):
        task = self.project.activity(name="Task - Basic Table")
        grid_widget = list(task.widgets())[0]

        associated_parts = grid_widget.associated_parts()
        self.assertTrue(associated_parts)
        self.assertIsInstance(associated_parts, tuple)
        self.assertTrue(len(associated_parts) == 2)

        models, instances = associated_parts

        self.assertIsInstance(models, list)
        self.assertTrue(all(isinstance(p, Part) for p in models))
        self.assertTrue(all(p.category == Category.MODEL for p in models))
        self.assertTrue(all(p.category == Category.INSTANCE for p in models))

    def test_bulk_update(self):
        parent = self.project.activity(name="Tasks with Widgets")
        task = self.project.activity(name="Task - Form + Tables + Service")
        test_task = task.clone(
            parent=parent,
            name="BULK WIDGET TEST TASK",
        )
        self.new_widget = test_task  # for teardown

        widgets = list(test_task.widgets())[1:]  # remove meta-panel
        self.client.update_widgets(
            widgets=[dict(id=w.id, title="widget {}".format(i + 1)) for i, w in enumerate(widgets)]
        )

        with self.assertRaises(IllegalArgumentError):
            double_update = [widgets[0], widgets[0], widgets[1]]
            self.client.update_widgets(
                widgets=[dict(id=w.id, title="widget {}".format(i + 1)) for i, w in enumerate(double_update)]
            )


class TestWidgetManager(TestBetamax):

    def setUp(self):
        super().setUp()
        self.wm = self.project.activity('Task - Form + Tables + Service').widgets()  # type: WidgetsManager

    def test_widget_manager_creation(self):
        with self.assertRaises(IllegalArgumentError):
            WidgetsManager(widgets=[], activity='Run away!')

    def test_widget_not_found(self):
        with self.assertRaises(NotFoundError):
            widget = self.wm['unladen swallow']

    def test_activity_has_metapanel_in_widget_manager(self):
        self.assertIsInstance(self.wm, WidgetsManager)
        self.assertTrue(len(self.wm) >= 5)

        self.assertTrue(self.wm[0].widget_type, WidgetTypes.METAPANEL)

    def test_widget_can_be_found_with_uuid_in_widget_manager(self):
        meta_panel = self.wm[0]

        self.assertTrue(meta_panel, WidgetTypes.METAPANEL)
        self.assertEqual(self.wm[meta_panel.id], meta_panel)

    def test_widget_in_widget_manager(self):
        for key in [0, self.wm[0], self.wm[0].id]:
            with self.subTest(msg='key: {} {}'.format(type(key), key)):
                widget = self.wm[key]
                self.assertIn(widget, self.wm)

    def test_widgetmanager_has_activity_and_client(self):
        self.assertIsNotNone(self.wm._client)
        self.assertIsInstance(self.wm.activity, Activity)
        self.assertIsInstance(self.wm._client, self.client.__class__)


class TestWidgetManagerInActivity(TestBetamax):
    def setUp(self):
        super(TestWidgetManagerInActivity, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.wm = self.task.widgets()  # type: WidgetsManager

    def tearDown(self):
        self.task.delete()
        super(TestWidgetManagerInActivity, self).tearDown()

    def test_new_widget_using_widget_manager(self):
        self.assertEqual(len(self.wm), 1)
        metapanel = self.wm[0]

        htmlwidget = self.wm.create_widget(
            widget_type=WidgetTypes.HTML,
            title="Test HTML widget",
            meta=dict(htmlContent="Hello")
        )

        self.assertIsInstance(metapanel, MetapanelWidget)
        self.assertIsInstance(htmlwidget, HtmlWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_edit_widget_meta(self):
        bike_part = self.project.part('Bike')
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )

        widget.meta.update({"showColumns": [ShowColumnTypes.UNIT, ShowColumnTypes.DESCRIPTION]})
        widget.edit(meta=widget.meta)

        live_widget = self.client.widget(pk=widget.id)

        self.assertEqual(widget.meta, live_widget.meta)

    def test_edit_widget_title(self):
        bike_part = self.project.part('Bike')
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )

        new_title = "My customly edited title"
        widget.edit(title=new_title)

        live_widget = self.client.widget(pk=widget.id)

        self.assertEqual(new_title, live_widget.title)
        self.assertEqual(new_title, live_widget.title_visible)

    def test_edit_widget_title_is_none(self):
        bike_part = self.project.part('Bike')
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )

        widget.edit(title=None)

        live_widget = self.client.widget(pk=widget.id)

        self.assertFalse(live_widget.title)
        self.assertIsNone(live_widget.title_visible)
        self.assertEqual(WidgetTitleValue.NO_TITLE, live_widget.meta.get("showTitleValue"))

        widget.edit(title=False)

        live_widget = self.client.widget(pk=widget.id)

        self.assertFalse(live_widget.title)
        self.assertEqual("Bike", live_widget.title_visible)
        self.assertEqual(WidgetTitleValue.DEFAULT, live_widget.meta.get("showTitleValue"))

    def test_edit_widget_title_and_meta(self):
        bike_part = self.project.part('Bike')
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )
        new_title = "New title"
        new_meta = dict(widget.meta)
        new_meta.update({"showColumns": [ShowColumnTypes.UNIT, ShowColumnTypes.DESCRIPTION]})
        widget.edit(
            title=new_title,
            meta=new_meta,
        )

        live_widget = self.client.widget(pk=widget.id)

        self.assertEqual(new_title, live_widget.title)
        self.assertEqual(new_title, live_widget.title_visible)
        self.assertEqual([ShowColumnTypes.UNIT, ShowColumnTypes.DESCRIPTION], live_widget.meta["showColumns"])

    def test_widget_title(self):
        title = 'Hidden title'
        bike_part = self.project.part('Bike')
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
            title=title,
            show_title_value=WidgetTitleValue.NO_TITLE,
        )

        self.assertEqual(title, widget.title)
        self.assertNotEqual(widget.title_visible, widget.title)
        self.assertIsNone(widget.title_visible)
        self.assertEqual(slugify_ref(title), widget.ref)
        self.assertEqual(WidgetTitleValue.NO_TITLE, widget.meta.get('showTitleValue'))

    def test_property_grid_with_associations_using_widget_manager(self):

        frame = self.project.part(name='Frame')
        frame_model = self.project.model(name='Frame')

        widget = self.wm.create_widget(
            widget_type=WidgetTypes.PROPERTYGRID,
            title="Frame Property Grid",
            meta=dict(
                activityId=str(self.task.id),
                partInstanceId=str(frame.id)
            ),
            writable_models=frame_model.properties,
            readable_models=[]
        )

        self.assertIsInstance(widget, PropertygridWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_attachment_widget(self):
        picture_instance = self.project.part('Bike').property('Picture')
        widget = self.wm.add_attachmentviewer_widget(
            attachment_property=picture_instance,
        )

        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual("Picture", widget.title_visible)

    def test_add_attachment_widget_with_associations_using_widget_manager(self):
        photo_property = self.project.property("Picture")

        widget = self.wm.add_attachmentviewer_widget(
            title="Attachment Viewer",
            attachment_property=photo_property,
        )

        self.assertEqual("Attachment Viewer", widget.title_visible)
        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_attachment_widget_with_editable_association(self):
        photo_property = self.project.property("Picture")

        widget = self.wm.add_attachmentviewer_widget(
            title="Attachment Viewer",
            attachment_property=photo_property,
            editable=True,
        )

        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_super_grid_widget(self):
        part_model = self.project.model(name='Wheel')
        parent_instance = self.project.part(name='Bike')

        widget = self.wm.add_supergrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=False,
            emphasize_edit=True,
            all_readable=True,
            incomplete_rows=True
        )

        self.assertIsInstance(widget, SupergridWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual(part_model.name, widget.title_visible)

    def test_add_filtered_grid_widget(self):
        part_model = self.project.model(name='Wheel')
        parent_instance = self.project.part(name='Bike')
        widget = self.wm.add_filteredgrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=True,
            # sort_property=part_model.property(name='Diameter'),
            emphasize_edit=True,
            all_writable=True,
            collapse_filters=False,
        )

        self.assertIsInstance(widget, FilteredgridWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual(part_model.name, widget.title_visible)

    def test_add_filtered_grid_widget_with_prefilters_and_excluded_propmodels(self):
        part_model = self.project.model(name='Wheel')
        parent_instance = self.project.part(name='Bike')
        excluded_propmodels = [part_model.property(name='Spokes')]
        prefilters = dict(
            property_models=[part_model.property(name='Diameter')],
            values=[66],
            filters_type=[FilterType.LOWER_THAN_EQUAL]
        )
        widget = self.wm.add_filteredgrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=True,
            sort_name=True,
            emphasize_edit=True,
            all_writable=True,
            collapse_filters=False,
            prefilters=prefilters,
            excluded_propmodels=excluded_propmodels
        )

        self.assertIsInstance(widget, FilteredgridWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_propertygrid_widget(self):
        bike_part = self.project.part(name='Bike')
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            show_headers=False, show_columns=[ShowColumnTypes.UNIT],
            readable_models=bike_part.model().properties[:2],
            writable_models=bike_part.model().properties[3:],
        )

        self.assertIsInstance(widget, PropertygridWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual(bike_part.name, widget.title_visible)

    def test_add_signature_widget(self):
        bike_part = self.project.part(name='Bike')
        picture = bike_part.property(name='Picture')

        widget1 = self.wm.add_signature_widget(attachment_property=picture, title="Yes, my precious",
                                               custom_undo_button_text="Remove za widget",
                                               custom_button_text="Sign za widget")
        widget2 = self.wm.add_signature_widget(attachment_property=picture, title=False,
                                               custom_undo_button_text=False,
                                               custom_button_text=False)
        widget3 = self.wm.add_signature_widget(attachment_property=picture)
        self.assertIsInstance(widget1, SignatureWidget)
        self.assertIsInstance(widget2, SignatureWidget)
        self.assertIsInstance(widget3, SignatureWidget)
        self.assertEqual(len(self.wm), 1 + 3)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_signature_widget(attachment_property='Failed script')

    def test_add_card_widget(self):
        # setUp
        bike_part = self.project.part(name='Bike')
        picture = bike_part.property(name='Picture')
        process = self.project.activities(activity_type=ActivityType.PROCESS)[0]

        widget1 = self.wm.add_card_widget(description='Some description')
        widget2 = self.wm.add_card_widget(image=picture, title=False, image_fit=ImageFitValue.COVER,
                                          link=self.task.id, link_target=LinkTargets.SAME_TAB)
        widget3 = self.wm.add_card_widget(title='Tree view', description='Process opens in tree view',
                                          link=process)

        # testing
        self.assertEqual(len(self.wm), 1 + 3)

        self.assertEqual(self.project.name, widget1.title_visible)
        self.assertIsInstance(widget1, CardWidget)
        self.assertIsNone(widget1.meta.get('customLink'))

        self.assertIsInstance(widget2, CardWidget)
        self.assertEqual(ImageFitValue.COVER, widget2.meta.get('imageFit'))
        self.assertEqual(CardWidgetLinkTarget.SAME_TAB, widget2.meta.get('linkTarget'))

        self.assertIsInstance(widget3, CardWidget)
        self.assertEqual(CardWidgetLinkValue.TREE_VIEW, widget3.meta.get('showLinkValue'))

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(title=12)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(description=bike_part)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(image='this should not work')

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(link=self.task, link_target='_somewhere')

        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.wm.add_card_widget(image_fit=3)

    def test_add_card_widget_ke_chain_pages(self):
        for native_page_name in KEChainPages.values():
            with self.subTest(msg='Page {}'.format(native_page_name)):
                card_widget = self.wm.add_card_widget(title=native_page_name, link=native_page_name)
                self.assertIsInstance(card_widget, CardWidget)

        self.assertEqual(len(self.wm), 9, msg='New KE-chain page has been added to the Enum, '
                                              'check if the mapping dicts in enums.py need updating too!')

    def test_add_service_widget(self):
        service_gears_successful = self.project.service("Service Gears - Successful")

        widget1 = self.wm.add_service_widget(service=service_gears_successful)
        widget2 = self.wm.add_service_widget(
            service=service_gears_successful,
            title=None,
            custom_button_text="Run this script (Custom!) and no title",
            emphasize_run=True,
            download_log=True,
        )
        widget3 = self.wm.add_service_widget(
            service=service_gears_successful,
            title="Also a custom title, but no log",
            custom_button_text="Run this script (Custom!)",
            emphasize_run=False,
            show_log=False,
        )

        self.assertIsInstance(widget1, ServiceWidget)
        self.assertIsInstance(widget2, ServiceWidget)
        self.assertIsInstance(widget3, ServiceWidget)
        self.assertEqual(len(self.wm), 1 + 3)
        self.assertEqual(service_gears_successful.name, widget1.title_visible)

    def test_add_html_widget(self):
        widget_1 = self.wm.add_html_widget(html='Or is this just fantasy?', title='Is this the real life?')

        self.assertIsInstance(widget_1, HtmlWidget)
        self.assertEqual(len(self.wm), 1 + 1)

        widget_2 = self.wm.add_html_widget(
            title='Caught in a landslide',
            show_title_value=WidgetTitleValue.NO_TITLE,
            html='No escape from reality',
        )

        self.assertEqual(widget_2.ref, slugify_ref('Caught in a landslide'))

    def test_add_metapanel_widget(self):
        widget = self.wm.add_metapanel_widget(
            show_all=True,
            show_progress=True,
            show_progressbar=True
        )

        self.assertEqual(len(self.wm), 1 + 1)
        self.assertIsNone(widget.title_visible)

    def test_add_metapanel_with_progress_settings(self):
        progress_bar = dict(
            height=15,
            showProgressText=False,
            colorCompleted=ProgressBarColors.RED
        )
        self.wm.add_metapanel_widget(
            show_all=False,
            show_progress=False,
            show_progressbar=True,
            progress_bar=progress_bar
        )
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_progress_widget(self):
        self.wm.add_progress_widget(custom_height=35,
                                    show_progress_text=False)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_delete_all_widgets(self):
        """Delete all widgets from an activity"""
        self.assertEqual(len(self.wm), 1)

        self.wm.delete_all_widgets()

        self.assertEqual(len(self.wm), 0)
        self.assertEqual(len(self.task.widgets()), 0)

    def test_delete_widget(self):
        """Delete single widget from an activity"""
        self.assertEqual(len(self.wm), 1)

        self.wm.delete_widget(0)

        self.assertEqual(len(self.wm), 0)
        self.assertEqual(len(self.task.widgets()), 0)

    def test_delete_widget_stand_alone(self):
        """Delete a widget by itself and check the manager"""
        widget_one = self.wm[0]
        widget_one.delete()

        self.assertNotIn(widget_one, self.wm)

    def test_add_notebook_widget(self):
        notebook = self.project.service(name="Service Gears - Successful")
        widget1 = self.wm.add_notebook_widget(
            notebook=notebook,
            title=False)

        widget2 = self.wm.add_notebook_widget(
            notebook=notebook,
            title="With custom title")

        widget3 = self.wm.add_notebook_widget(
            notebook=notebook.id,
            title="With no padding and custom height",
            customHeight=400,
            noPadding=True)

        self.assertIsInstance(widget1, NotebookWidget)
        self.assertIsInstance(widget2, NotebookWidget)
        self.assertIsInstance(widget3, NotebookWidget)
        self.assertEqual(len(self.wm), 1 + 3)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_notebook_widget(notebook="This will raise an error")

    def test_add_multicolumn_widget(self):
        bike_part = self.project.part('Bike')
        picture_instance = bike_part.property('Picture')
        picture_model = picture_instance.model()
        multi_column_widget = self.wm.add_multicolumn_widget(title="Multi column Grid + Attachment")

        widget1 = self.wm.add_propertygrid_widget(part_instance=bike_part,
                                                  writable_models=[picture_model],
                                                  parent_widget=multi_column_widget)

        widget2 = self.wm.add_attachmentviewer_widget(
            attachment_property=picture_instance, parent_widget=multi_column_widget.id
        )

        self.assertIsInstance(multi_column_widget, MulticolumnWidget)
        self.assertIsInstance(widget1, PropertygridWidget)
        self.assertIsInstance(widget2, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 3)

    def test_parent(self):
        multi_column_widget = self.wm.add_multicolumn_widget(title="Multi column widget")

        child_widget = self.wm.add_html_widget(
            title='Text widget', html='Text', parent_widget=multi_column_widget)

        parent_widget = child_widget.parent()

        self.assertEqual(multi_column_widget, parent_widget)

        with self.assertRaises(NotFoundError):
            parent_widget.parent()

    def test_add_scope_widget(self):
        scope_widget = self.wm.add_scope_widget()

        self.assertIsInstance(scope_widget, ScopeWidget)

    def test_scope_widget_invalid_inputs(self):
        for inputs in [
            dict(team='5'),
            dict(add=1),
            dict(edit=0),
            dict(emphasize_add='True'),
            dict(emphasize_edit=1.0),
            dict(show_columns='All', show_all_columns=False),
            dict(show_columns=['Any'], show_all_columns=False),
            dict(show_all_columns=0),
            dict(page_size=0),
            dict(tags='one'),
            dict(tags=['one', 2, 'three']),
            dict(sorted_column='Project name'),
            dict(sorted_direction='Alphabetical'),
            dict(active_filter='Active'),
            dict(search_filter='Project'),
        ]:
            with self.subTest(msg=inputs):
                with self.assertRaises(IllegalArgumentError):
                    self.wm.add_scope_widget(**inputs)

    def test_insert_widget(self):
        bike_part = self.project.part('Bike')
        w0 = self.wm[0]  # meta panel

        w1, w2, w3 = [self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
            title='Original widget {i} (w{i})'.format(i=i + 1),
        ) for i in range(3)]

        # if widget order is `[w0,w1,w2]` and inserting `w3` at index 1 (before Widget1);
        #           index:     0 ^1  2
        # the list will be `[w0,w3,w1,w2]`
        self.wm.insert(1, w3)

        self.assertTrue(w0.widget_type, WidgetTypes.METAPANEL)
        self.assertEqual(self.wm[w0.id].order, 0)
        self.assertEqual(self.wm[w3.id].order, 1)
        self.assertEqual(self.wm[w1.id].order, 2)
        self.assertEqual(self.wm[w2.id].order, 3)

        added_widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
            title="Widget Finally Positioned Under Metapanel",
            order=1,
        )

        self.assertEqual(added_widget.order, 1)
        self.assertEqual(self.wm[w3.id].order, 2)
        self.assertEqual(self.wm[w1.id].order, 3)
        self.assertEqual(self.wm[w2.id].order, 4)

    def test_create_widgets(self):
        # setUp
        new_widgets = self.wm.create_widgets(widgets=[
            dict(
                widget_type=WidgetTypes.HTML,
                title='A new text widget',
                meta=dict(
                    htmlContent='This is HTML text.'
                ),
            ),
            dict(
                widget_type=WidgetTypes.HTML,
                title='Another HTML widget',
                meta=dict(
                    htmlContent='They keep on multiplying.'
                ),
            ),
        ])

        # testing
        self.assertIsInstance(new_widgets, list)
        self.assertTrue(all(isinstance(w, Widget) for w in new_widgets))
        self.assertEqual(len(new_widgets), 2)

        # tearDown
        [w.delete() for w in new_widgets]

    def test_add_service_card_widget(self):
        # setUp
        service_gears_successful = self.project.service("Service Gears - Successful")
        bike_part = self.project.part(name='Bike')
        picture = bike_part.property(name='Picture')
        title = 'Service card title'
        description = 'The description'

        widget = self.wm.add_service_card_widget(
            service=service_gears_successful,
            image=picture,
            title=title,
            description=description,
            custom_button_text=False,
            link=KEChainPages.EXPLORER,
            link_target=LinkTargets.SAME_TAB,
            image_fit=ImageFitValue.COVER
        )

        self.assertEqual(len(self.wm), 1 + 1)

        self.assertIsInstance(widget, ServicecardWidget)
        self.assertEqual(widget.title, title)
        self.assertEqual(widget.meta.get('customText'), service_gears_successful.name)
        self.assertEqual(widget.meta.get('customDescription'), description)
        self.assertIn(picture.id, widget.meta.get('customImage'))
        self.assertEqual(ImageFitValue.COVER, widget.meta.get('imageFit'))
        self.assertEqual(CardWidgetLinkTarget.SAME_TAB, widget.meta.get('linkTarget'))

    def test_add_dashboard_widget(self):
        # setUp
        title_widget_1 = 'Dashboard widget'
        title_widget_2 = 'Tagged projects widget'

        widget_current_project = self.wm.add_dashboard_widget(
            title=title_widget_1,
            show_assignees=True
        )
        widget_tagged_projects = self.wm.add_dashboard_widget(
            title=title_widget_2,
            source_scopes_tags=['catalog-scope'],
            show_tasks=[DashboardWidgetShowTasks.UNASSIGNED_TASKS, DashboardWidgetShowTasks.CLOSED_TASKS],
            show_scopes=[DashboardWidgetShowScopes.CLOSED_SCOPES],
            show_assignees=False
        )

        # testing
        self.assertIsInstance(widget_current_project, DashboardWidget)
        self.assertIsInstance(widget_tagged_projects, DashboardWidget)

        self.assertEqual(widget_current_project.title, title_widget_1)
        self.assertEqual(widget_tagged_projects.title, title_widget_2)

        self.assertTrue(all(elem['selected'] for elem in widget_current_project.meta['showNumbers']))
        self.assertTrue(all(elem['selected'] for elem in widget_current_project.meta['showNumbersProjects']))

        self.assertTrue(any(elem['selected'] is False for elem in widget_tagged_projects.meta['showNumbers']))
        self.assertTrue(any(elem['selected'] is False for elem in widget_tagged_projects.meta['showNumbersProjects']))

        self.assertTrue(widget_current_project.meta['showAssignees'])
        self.assertFalse(widget_tagged_projects.meta['showAssignees'])
        return


class TestWidgetManagerWeatherWidget(TestBetamax):
    def setUp(self):
        super(TestWidgetManagerWeatherWidget, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.wm = self.task.widgets()  # type: WidgetsManager

        catalog_root_model = self.project.part(name='Catalog', classification=Classification.CATALOG,
                                               category=Category.MODEL)
        self.part_model_with_weather_prop = self.project.create_model_with_properties(
            parent=catalog_root_model,
            name='___TEST PART', multiplicity=Multiplicity.ONE,
            properties_fvalues=[
                dict(name='weather',
                     property_type=PropertyType.WEATHER_VALUE)])
        self.weather_prop_instance = self.part_model_with_weather_prop.instances()[0].property('weather')

    def tearDown(self):
        self.weather_widget.delete()
        self.part_model_with_weather_prop.delete()
        self.task.delete()
        super(TestWidgetManagerWeatherWidget, self).tearDown()

    def test_weather_widget(self):
        """Testing the weather widget."""
        self.weather_widget = self.wm.add_weather_widget(
            weather_property=self.weather_prop_instance,
        )


class TestWidgetNavigationBarWidget(TestBetamax):

    def setUp(self):
        super(TestWidgetNavigationBarWidget, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.wm = self.task.widgets()  # type: WidgetsManager

        self.frame = self.project.part(name='Frame')
        self.frame_model = self.project.model(name='Frame')

        activity_1 = self.project.activity('Task - Basic Table')
        activity_2 = self.project.activity('Task - Form')
        activity_3 = self.project.activity('Task - Service Increase Gears')

        self.nav_bar_config = [
            {'activityId': activity_1, 'emphasize': True},
            {'activityId': activity_2, 'customText': 'Hit me baby!'},
            {'activityId': activity_3}
        ]

    def tearDown(self):
        self.task.delete()
        super(TestWidgetNavigationBarWidget, self).tearDown()

    def test_add_navbar_widget(self):
        widget = self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            title="Navbar",
            alignment=Alignment.LEFT
        )

        self.assertIsInstance(widget, TasknavigationbarWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_navbar_widget_in_place(self):
        """
        Test whether the inputs to the function are updated in-place (incorrect) or not (correct).
        For example, activity objects are replaced by their IDs and customText is converted to a string.
        """
        self.nav_bar_config[0]['customText'] = False

        import copy
        original_bar = copy.deepcopy(self.nav_bar_config)

        self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=Alignment.LEFT,
        )

        self.assertEqual(original_bar, self.nav_bar_config)

    def test_add_navbar_widget_incorrect_keys(self):
        self.nav_bar_config[0]['emphasizeButton'] = True

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_tasknavigationbar_widget(
                activities=self.nav_bar_config,
                alignment=Alignment.LEFT,
            )

    def test_add_navbar_external_link(self):
        link = 'https://www.google.com'

        button_1 = self.nav_bar_config[0]
        button_1.pop('activityId')
        button_1['link'] = link

        widget = self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=Alignment.LEFT,
        )

        self.assertEqual(widget.meta['taskButtons'][0]['link'], link)

    def test_add_navbar_disabled_button(self):
        button_1 = self.nav_bar_config[0]
        button_1['isDisabled'] = True

        widget = self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=Alignment.LEFT,
        )

        self.assertTrue(widget.meta['taskButtons'][0]['isDisabled'])


class TestWidgetsCopyMove(TestBetamax):
    def setUp(self):
        super(TestWidgetsCopyMove, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.task_2 = self.project.create_activity(name="test_copy_widget")  # type: Activity

    def tearDown(self):
        self.task.delete()
        self.task_2.delete()
        super(TestWidgetsCopyMove, self).tearDown()

    def test_copy_widget(self):
        # setUp
        title = "Widget to copy"
        bike_part = self.project.part('Bike')
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(part_instance=bike_part, all_writable=True,
                                                          title=title)
        widget_1.copy(target_activity=self.task_2, order=0)
        widget_manager_2 = self.task_2.widgets()  # type: WidgetsManager
        associated_model = widget_manager_2[0].parts(category=Category.MODEL)[0]

        # testing
        self.assertEqual(widget_manager_2[0].widget_type, WidgetTypes.PROPERTYGRID)
        self.assertEqual(widget_manager_2[0].title, title)
        self.assertTrue(all(prop.output for prop in associated_model.properties))
        self.assertEqual(len(self.task.widgets()), 2)

    def test_copy_widget_with_wrong_target(self):
        # setUp
        title = "Widget to copy"
        bike_part = self.project.part('Bike')
        self.wm = self.task.widgets()  # type: WidgetsManager
        widget_1 = self.wm.add_propertygrid_widget(part_instance=bike_part, all_writable=True,
                                                   title=title)

        # testing
        with self.assertRaises(IllegalArgumentError):
            widget_1.copy(target_activity=self.project, order=0)

    def test_move_widget(self):
        # setUp
        title = "Widget to move"
        bike_part = self.project.part('Bike')
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(part_instance=bike_part, all_writable=True,
                                                          title=title)
        widget_1.move(target_activity=self.task_2, order=0)
        widget_manager_2 = self.task_2.widgets()  # type: WidgetsManager
        associated_model = widget_manager_2[0].parts(category=Category.MODEL)[0]

        # testing
        self.assertEqual(widget_manager_2[0].widget_type, WidgetTypes.PROPERTYGRID)
        self.assertEqual(widget_manager_2[0].title, title)
        self.assertTrue(all(prop.output for prop in associated_model.properties))
        self.assertEqual(1, len(self.task.widgets()))


class TestWidgetDownloadAsExcel(TestBetamax):

    def setUp(self):
        super().setUp()
        task = self.project.activity(name="Task - Form + Tables + Service")
        all_widgets = task.widgets()  # type: List[Widget]

        self.grid_widget = find(all_widgets, lambda w: w.widget_type == WidgetTypes.FILTEREDGRID)
        self.form_widget = find(all_widgets, lambda w: w.widget_type == WidgetTypes.PROPERTYGRID)

    def tearDown(self):
        super().tearDown()

    def test_download_as_excel(self):
        file_name = "My excel file"

        with temp_chdir() as target_dir:

            path = self.grid_widget.download_as_excel(
                target_dir=target_dir,
                file_name=file_name,
            )

            self.assertTrue(os.path.exists(path))
            self.assertIn(slugify_ref(file_name), path)

    def test_timezone_aware(self):
        file_name = "My excel file.xls"
        user = self.client.user("testuser")
        timezone_definition = user._json_data['timezone']

        self.assertEqual("Europe/Amsterdam", timezone_definition)

        with temp_chdir() as target_dir:
            path = self.grid_widget.download_as_excel(
                target_dir=target_dir,
                file_name=file_name,
                user=user,
            )

            self.assertTrue(os.path.exists(path))

    # noinspection PyTypeChecker
    def test_invalid_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.form_widget.download_as_excel()

        with self.assertRaises(IllegalArgumentError):
            self.grid_widget.download_as_excel(target_dir=3)

        with self.assertRaises(IllegalArgumentError):
            self.grid_widget.download_as_excel(file_name=15.3)

        with self.assertRaises(IllegalArgumentError):
            self.grid_widget.download_as_excel(user="Testuser")



import datetime
import os
import time
from typing import List, Text, Dict, Optional, Union
from urllib.parse import urljoin

import requests

from pykechain.defaults import ASYNC_REFRESH_INTERVAL, ASYNC_TIMEOUT_LIMIT, API_EXTRA_PARAMS
from pykechain.enums import ActivityType, ActivityStatus, Category, ActivityClassification, ActivityRootNames, \
    PaperSize, PaperOrientation
from pykechain.exceptions import NotFoundError, IllegalArgumentError, APIError, MultipleFoundError
from pykechain.models.activity2 import Activity2
from pykechain.models.input_checks import check_datetime, check_text, check_list_of_text, check_enum, check_user, \
    check_type, check_base
from pykechain.models.representations.component import RepresentationsComponent
from pykechain.models.tags import TagsMixin
from pykechain.models.tree_traversal import TreeObject
from pykechain.models.user import User
from pykechain.models.widgets.widgets_manager import WidgetsManager
from pykechain.utils import parse_datetime, is_valid_email, empty, Empty, clean_empty_values


class Activity(TreeObject, TagsMixin, Activity2):
    """A virtual object representing a KE-chain activity.

    .. versionadded:: 2.0

    :ivar id: id of the activity
    :type id: uuid
    :ivar name: name of the activity
    :type name: basestring
    :ivar created_at: created datetime of the activity
    :type created_at: datetime
    :ivar updated_at: updated datetime of the activity
    :type updated_at: datetime
    :ivar description: description of the activity
    :type description: basestring
    :ivar status: status of the activity. One of :class:`pykechain.enums.ActivityStatus`
    :type status: basestring
    :ivar classification: classification of the activity. One of :class:`pykechain.enums.ActivityClassification`
    :type classification: basestring
    :ivar activity_type: Type of the activity. One of :class:`pykechain.enums.ActivityType` for WIM version 2
    :type activity_type: basestring
    """

    def __init__(self, json, **kwargs):
        """Construct an Activity from a json object."""
        super().__init__(json, **kwargs)

        self._scope_id = json.get('scope_id')

        self.ref = json.get('ref')  # type: Text
        self.description = json.get('description', '')  # type: Text
        self.status = json.get('status')  # type: ActivityStatus
        self.classification = json.get('classification')  # type: ActivityClassification
        self.activity_type = json.get('activity_type')  # type: ActivityType
        self.start_date = parse_datetime(json.get('start_date'))
        self.due_date = parse_datetime(json.get('due_date'))
        self.assignees_ids = json.get('assignees_ids', [])  # type: List[Text]
        self._options = json.get('activity_options', {})

        self._tags = json.get('tags', [])  # type: List[Text]
        self._representations_container = RepresentationsComponent(
            self,
            self._options.get('representations', {}),
            self._save_representations,
        )
        self._widgets_manager = None  # type: Optional[WidgetsManager]

    def __call__(self, *args, **kwargs) -> 'Activity':
        """Short-hand version of the `child` method."""
        return self.child(*args, **kwargs)

    def refresh(self, *args, **kwargs):
        """Refresh the object in place."""
        super().refresh(url=self._client._build_url('activity', activity_id=self.id),
                        extra_params=API_EXTRA_PARAMS['activity'], *args, **kwargs)

    #
    # additional properties
    #

    @property
    def assignees(self) -> List['User']:
        """List of assignees to the activity.

        Provides a list of `User` objects or an empty list.

        :return: a list of `User` objects or an empty list.
        :rtype: list
        """
        return self._client.users(id__in=self.assignees_ids, is_hidden=False) if self.assignees_ids else []

    @property
    def scope_id(self):
        """
        Id of the scope this Activity belongs to.

        This property will always produce a scope_id, even when the scope object was not included in an earlier
        response.

        When the :class:`Scope` is not included in this task, it will make an additional call to the KE-chain API.

        :return: the scope id
        :type: uuid
        :raises NotFoundError: if the scope could not be found
        """
        if self._scope_id is None:
            self.refresh()
            if self._scope_id is None:
                raise NotFoundError("Activity '{}' has no related scope!".format(self))
        return self._scope_id

    @scope_id.setter
    def scope_id(self, value):
        self._scope_id = value

    @property
    def representations(self):
        """Get and set the activity representations."""
        return self._representations_container.get_representations()

    @representations.setter
    def representations(self, value):
        self._representations_container.set_representations(value)

    def _save_representations(self, representation_options):
        self._options.update({'representations': representation_options})
        self.edit(activity_options=self._options)

    #
    # predicates
    #

    def is_rootlevel(self):
        """
        Determine if the Activity is at the toplevel of a project, i.e. below the root itself.

        It will look for the name of the parent which should be either ActivityRootNames.WORKFLOW_ROOT or
        ActivityRootNames.CATALOG_ROOT. If the name of the parent cannot be found an additional API call is made
        to retrieve the parent object (based on the `parent_id` in the json_data).

        :return: Return True if it is a root level activity, otherwise return False
        :rtype: bool
        """
        # when the activity itself is a root, than return False immediately
        if self.is_root():
            return False

        parent_name = None
        parent_dict = self._json_data.get('parent_id_name')

        if parent_dict and 'name' in parent_dict:
            parent_name = parent_dict.get('name')
        if not parent_dict:
            parent_name = self._client.activity(id=self.parent_id).name

        return parent_name in ActivityRootNames.values()

    def is_task(self):
        """
        Determine if the Activity is of ActivityType.TASK.

        :return: Return True if it is a task, otherwise return False
        :rtype: bool
        """
        return self.activity_type == ActivityType.TASK

    def is_subprocess(self):
        """
        Determine if the Activity is of ActivityType.PROCESS.

        :return: Return True if it is a subprocess, otherwise return False
        :rtype: bool
        """
        return self.is_process()

    def is_process(self):
        """
        Determine if the Activity is of ActivityType.PROCESS.

        :return: Return True if it is a process, otherwise return False
        :rtype: bool
        """
        return self.activity_type == ActivityType.PROCESS

    def is_workflow(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.WORKFLOW.

        :return: Return True if it is a workflow classification activity, otherwise return False
        :rtype: bool
        """
        return self.classification == ActivityClassification.WORKFLOW

    def is_catalog(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.CATALOG.

        :return: Return True if it is a catalog classification activity, otherwise return False
        :rtype: bool
        """
        return self.classification == ActivityClassification.CATALOG

    def is_workflow_root(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.WORKFLOW and a ROOT object.

        :return: Return True if it is a root workflow classification activity, otherwise return False
        :rtype: bool
        """
        return self.is_root() and self.is_workflow()

    def is_catalog_root(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.CATALOG and a ROOT object.

        :return: Return True if it is a root catalog classification activity, otherwise return False
        :rtype: bool
        """
        return self.is_root() and self.is_catalog()

    def is_root(self):
        """
        Determine if the Activity is a ROOT object.

        If you want to determine if it is also a workflow or a catalog root, use :func:`Activity.is_workflow_root()`
        or :func:`Activity.is_catalog_root()` methods.

        :return: Return True if it is a root object, otherwise return False
        :rtype: bool
        """
        return self.name in ActivityRootNames.values() and self.parent_id is None

    def is_configured(self):
        """
        Determine if the Activity is configured with input and output properties.

        Makes an additional lightweight call to the API to determine if any associated models are there.

        :return: Return True if it is configured, otherwise return False
        :rtype: bool
        """
        # check configured based on if we get at least 1 part back
        return bool(self.parts(category=Category.MODEL, limit=1))

    def is_customized(self):
        """
        Determine if the Activity is customized.

        In other words if it has a customization. Use can use the :func:`Activity.customization()` to retrieve
        the customization object and configure the task.

        :return: Return True if it is customized, otherwise return False
        :rtype: bool
        """
        return bool(self._json_data.get('customization', False))

    #
    # methods
    #

    def create(self, *args, **kwargs) -> 'Activity':
        """Create a new activity belonging to this subprocess.

        See :func:`pykechain.Client.create_activity` for available parameters.

        :raises IllegalArgumentError: if the `Activity` is not a `PROCESS`.
        :raises APIError: if an Error occurs.
        """
        if self.activity_type != ActivityType.PROCESS:
            raise IllegalArgumentError("One can only create a task under a subprocess.")
        return self._client.create_activity(self, *args, **kwargs)

    def parent(self) -> 'Activity':
        """Retrieve the parent in which this activity is defined.

        If this is a task on top level, it raises NotFounderror.

        :return: a :class:`Activity`
        :raises NotFoundError: when it is a task in the top level of a project
        :raises APIError: when other error occurs

        Example
        -------
        >>> task = project.activity('Subtask')
        >>> parent_of_task = task.parent()

        """
        if self.parent_id is None:
            raise NotFoundError("Cannot find parent for task '{}', as this task exist on top level.".format(self))
        elif self._parent is None:
            self._parent = self._client.activity(pk=self.parent_id, scope=self.scope_id)
        return self._parent

    def children(self, **kwargs) -> List['Activity']:
        """Retrieve the direct activities of this subprocess.

        It returns a combination of Tasks (a.o. UserTasks) and Subprocesses on the direct descending level.
        Only when the activity is a Subprocess, otherwise it raises a NotFoundError

        :param kwargs: Additional search arguments, check :func:`pykechain.Client.activities` for additional info
        :return: a list of :class:`Activity`
        :raises NotFoundError: when this task is not of type `ActivityType.PROCESS`

        Example
        -------
        >>> task = project.activity('Subprocess')
        >>> children = task.children()

        Example searching for children of a subprocess which contains a name (icontains searches case insensitive)

        >>> task = project.activity('Subprocess')
        >>> children = task.children(name__icontains='more work')

        """
        if self.activity_type != ActivityType.PROCESS:
            raise NotFoundError("Only subprocesses can have children, please choose a subprocess instead of a '{}' "
                                "(activity '{}')".format(self.activity_type, self.name))
        if not kwargs:
            if self._cached_children is None:
                self._cached_children = self._client.activities(parent_id=self.id, scope=self.scope_id, **kwargs)
                for child in self._cached_children:
                    child._parent = self
            return self._cached_children
        else:
            return self._client.activities(parent_id=self.id, scope=self.scope_id, **kwargs)

    def child(self,
              name: Optional[Text] = None,
              pk: Optional[Text] = None,
              **kwargs) -> 'Activity':
        """
        Retrieve a child object.

        :param name: optional, name of the child
        :type name: str
        :param pk: optional, UUID of the child
        :type: pk: str
        :return: Child object
        :raises MultipleFoundError: whenever multiple children fit match inputs.
        :raises NotFoundError: whenever no child matching the inputs could be found.
        """
        if not (name or pk):
            raise IllegalArgumentError('You need to provide either "name" or "pk".')

        if self._cached_children:
            # First try to find the child without calling KE-chain.
            if name:
                activity_list = [c for c in self.children() if c.name == name]
            else:
                activity_list = [c for c in self.children() if c.id == pk]
        else:
            activity_list = []

        if not activity_list:
            if name:
                activity_list = self.children(name=name)
            else:
                activity_list = self.children(pk=pk)

        criteria = '\nname: {}\npk: {}\nkwargs: {}'.format(name, pk, kwargs)

        if len(activity_list) == 1:
            child = activity_list[0]

        elif len(activity_list) > 1:
            raise MultipleFoundError('{} has more than one matching child.{}'.format(self, criteria))
        else:
            raise NotFoundError('{} has no matching child.{}'.format(self, criteria))
        return child

    def siblings(self, **kwargs) -> List['Activity']:
        """Retrieve the other activities that also belong to the parent.

        It returns a combination of Tasks (a.o. UserTasks) and Subprocesses on the level of the current task, including
        itself. This also works if the activity is of type `ActivityType.PROCESS`.

        :param kwargs: Additional search arguments, check :func:`pykechain.Client.activities` for additional info
        :return: list of :class:`Activity`
        :raises NotFoundError: when it is a task in the top level of a project

        Example
        -------
        >>> task = project.activity('Some Task')
        >>> siblings = task.siblings()

        Example for siblings containing certain words in the task name
        >>> task = project.activity('Some Task')
        >>> siblings = task.siblings(name__contains='Another Task')

        """
        if self.parent_id is None:
            raise NotFoundError("Cannot find siblings for task '{}', as this task exist on top level.".format(self))
        return self._client.activities(parent_id=self.parent_id, scope=self.scope_id, **kwargs)

    def all_children(self) -> List['Activity']:
        """
        Retrieve a flat list of all descendants, sorted depth-first.

        Returns an empty list for Activities of type TASK.

        :returns list of child objects
        :rtype List
        """
        if self.activity_type == ActivityType.TASK:
            return []
        return super().all_children()

    def count_children(self, **kwargs) -> int:
        """
        Retrieve the number of child activities using a light-weight request.

        :return: number of Activities
        :rtype int
        """
        if self.activity_type != ActivityType.PROCESS:
            raise IllegalArgumentError(
                "You can only count the number of children of an Activity of type {}".format(ActivityType.PROCESS))

        return super().count_children(method="activities")

    def clone(
        self,
        parent: Optional[Union['Activity', Text]] = None,
        update_dict: Optional[Dict] = None,
        **kwargs
    ) -> Optional['Activity']:
        """
        Create a copy of this activity.

        :param parent: (O) parent Activity object or UUID
        :type parent: Activity
        :param update_dict: (O) dictionary of new values to set on the cloned activities, e.g. `{"name": "New name"}`
        :type update_dict: dict
        :param kwargs: additional arguments, see the `Client.clone_activities()` method
        :return: clone of this activity
        :rtype Activity
        """
        update_dict = check_type(update_dict, dict, 'update_dict')
        if update_dict:
            validated_dict = self._validate_edit_arguments({}, **update_dict)
        else:
            validated_dict = None

        cloned_activities = self._client.clone_activities(
            activity_parent=check_base(parent, Activity, 'parent') or self.parent_id,
            activities=[self],
            activity_update_dicts={self.id: validated_dict} if validated_dict else None,
            **kwargs
        )
        return cloned_activities[0] if cloned_activities else None

    def edit_cascade_down(
            self,
            start_date: Optional[Union[datetime.datetime, Empty]] = empty,
            due_date: Optional[Union[datetime.datetime, Empty]] = empty,
            assignees: Optional[Union[List[Text], Empty]] = empty,
            assignees_ids: Optional[Union[List[Text], Empty]] = empty,
            status: Optional[Union[ActivityStatus, Text, Empty]] = empty,
            overwrite: Optional[bool] = False,
            **kwargs
    ) -> None:
        """
        Edit the activity and all its descendants with a single operation.

        :param start_date: (optionally) edit the start date of the activity as a datetime object (UTC time/timezone
                            aware preferred)
        :type start_date: datetime or None
        :param due_date: (optionally) edit the due_date of the activity as a datetime object (UTC time/timzeone
                            aware preferred)
        :type due_date: datetime or None
        :param assignees: (optionally) edit the assignees (usernames) of the activity as a list
        :type assignees: list(basestring) or None
        :param assignees_ids: (optionally) edit the assignees (user id's) of the activity as a list
        :type assignees_ids: list(basestring) or None
        :param status: (optionally) edit the status of the activity as a string based
                       on :class:`~pykechain.enums.ActivityStatus`
        :type status: ActivityStatus, basestring or None
        :param overwrite: (optionally) whether to overwrite existing assignees (True) or
                          merge with existing assignees (False, default)
        :type overwrite: bool

        :return: flat list of the current task all descendants that have been edited
        :rtype list[Activity]
        """
        update_dict = {'id': self.id}

        self._validate_edit_arguments(
            update_dict=update_dict,
            start_date=start_date,
            due_date=due_date,
            assignees=assignees,
            assignees_ids=assignees_ids,
            status=status,
            **kwargs
        )

        all_tasks = [self] + self.all_children()
        new_assignees = update_dict.get('assignees_ids', list())

        # Create update-json
        data = list()

        update_dict = clean_empty_values(update_dict=update_dict)

        for task in all_tasks:
            task_specific_update_dict = dict(update_dict)

            if not overwrite:
                # Append the existing assignees of the task to the new assignees
                existing_assignees = [u.id for u in task.assignees]
                task_specific_update_dict['assignees_ids'] = list(set(existing_assignees + new_assignees))

            task_specific_update_dict.update({'id': task.id})
            data.append(task_specific_update_dict)

        # Perform bulk update
        self._client.update_activities(activities=data)

    def edit(
            self,
            name: Optional[Union[Text, Empty]] = empty,
            description: Optional[Union[Text, Empty]] = empty,
            start_date: Optional[Union[datetime.datetime, Empty]] = empty,
            due_date: Optional[Union[datetime.datetime, Empty]] = empty,
            assignees: Optional[Union[List[Text], Empty]] = empty,
            assignees_ids: Optional[Union[List[Text], Empty]] = empty,
            status: Optional[Union[ActivityStatus, Text, Empty]] = empty,
            tags: Optional[Union[List[Text], Empty]] = empty,
            **kwargs
    ) -> None:
        """Edit the details of an activity.

        Setting an input to None will clear out the value (exception being name and status).

        :param name: (optionally) edit the name of the activity. Name cannot be cleared.
        :type name: basestring or None or Empty
        :param description: (optionally) edit the description of the activity or clear it
        :type description: basestring or None or Empty
        :param start_date: (optionally) edit the start date of the activity as a datetime object (UTC time/timezone
                            aware preferred) or clear it
        :type start_date: datetime or None or Empty
        :param due_date: (optionally) edit the due_date of the activity as a datetime object (UTC time/timzeone
                            aware preferred) or clear it
        :type due_date: datetime or None or Empty
        :param assignees: (optionally) edit the assignees (usernames) of the activity as a list, will overwrite all
                          assignees or clear them
        :type assignees: list(basestring) or None or Empty
        :param assignees_ids: (optionally) edit the assignees (user id's) of the activity as a list, will overwrite all
                             assignees or clear them
        :type assignees_ids: list(basestring) or None or Empty
        :param status: (optionally) edit the status of the activity as a string based. Status cannot be cleared.
                       on :class:`~pykechain.enums.ActivityStatus`
        :type status: ActivityStatus or None or Empty
        :param tags: (optionally) replace the tags on an activity, which is a list of strings ["one","two","three"]
                    or clear them
        :type tags: list of basestring or None or Empty

        :raises NotFoundError: if a `username` in the list of assignees is not in the list of scope members
        :raises IllegalArgumentError: if the type of the inputs is not correct
        :raises APIError: if another Error occurs
        :warns: UserWarning - When a naive datetime is provided. Defaults to UTC.

        Example
        -------
        >>> from datetime import datetime
        >>> my_task = project.activity('Specify the wheel diameter')
        >>> my_task.edit(name='Specify wheel diameter and circumference',
        ...              description='The diameter and circumference are specified in inches',
        ...              start_date=datetime.utcnow(),  # naive time is interpreted as UTC time
        ...              assignee='testuser')

        If we want to provide timezone aware datetime objects we can use the 3rd party convenience library :mod:`pytz`.
        Mind that we need to fetch the timezone first and use `<timezone>.localize(<your datetime>)` to make it
        work correctly.

        Using `datetime(2017,6,1,23,59,0 tzinfo=<tz>)` does NOT work for most timezones with a
        daylight saving time. Check the `pytz <http://pythonhosted.org/pytz/#localized-times-and-date-arithmetic>`_
        documentation.

        To make it work using :mod:`pytz` and timezone aware :mod:`datetime` see the following example::

        >>> import pytz
        >>> start_date_tzaware = datetime.now(pytz.utc)
        >>> mytimezone = pytz.timezone('Europe/Amsterdam')
        >>> due_date_tzaware = mytimezone.localize(datetime(2019, 10, 27, 23, 59, 0))
        >>> my_task.edit(due_date=due_date_tzaware, start_date=start_date_tzaware)

        Not mentioning an input parameter in the function will leave it unchanged. Setting a parameter as None will
        clear its value (where that is possible). The example below will clear the due_date, but leave everything else
        unchanged.

        >>> my_task.edit(due_date=None)

        """
        update_dict = {
            'id': self.id,
            'name': check_text(text=name, key='name') or self.name,
            'description': check_text(text=description, key='description') or str(),
            'tags': check_list_of_text(tags, 'tags', True) or list(),
        }

        self._validate_edit_arguments(
            update_dict=update_dict,
            start_date=start_date,
            due_date=due_date,
            assignees=assignees,
            assignees_ids=assignees_ids,
            status=status,
            **kwargs
        )

        update_dict = clean_empty_values(update_dict=update_dict)

        url = self._client._build_url('activity', activity_id=self.id)

        response = self._client._request('PUT', url, json=update_dict, params=API_EXTRA_PARAMS['activity'])

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Activity {}".format(self), response=response)

        self.refresh(json=response.json().get('results')[0])

    def _validate_edit_arguments(
            self,
            update_dict,
            start_date=None,
            due_date=None,
            assignees=None,
            assignees_ids=None,
            status=None,
            **kwargs
    ) -> Dict:
        """Verify inputs provided in both the `clone`, `edit` and `edit_cascade_down` methods."""
        update_dict.update({
            'start_date': check_datetime(dt=start_date, key='start_date'),
            'due_date': check_datetime(dt=due_date, key='due_date'),
            'status': check_enum(status, ActivityStatus, 'status') or self.status
        })

        # If both are empty that means the user is not interested in changing them
        if isinstance(assignees_ids, Empty) and isinstance(assignees, Empty):
            pass

        # If one of them is None, then the assignees will be cleared from the Activity
        elif assignees is None or assignees_ids is None:
            update_dict['assignees_ids'] = list()

        # In case both of them have values specified, then an Error is raised
        elif assignees and assignees_ids and not isinstance(assignees, Empty) and not isinstance(assignees_ids, Empty):
            raise IllegalArgumentError('Provide either assignee names or their ids, but not both.')
        # Otherwise, pick the one that has a value specified which is not Empty
        else:
            assignees = assignees if assignees is not None and not isinstance(assignees, Empty) else assignees_ids

            if assignees:
                if not isinstance(assignees, (list, tuple, set)) or \
                        not all(isinstance(a, (str, int)) for a in assignees):
                    raise IllegalArgumentError('All assignees must be provided as list, tuple or set of names or IDs.')

                update_assignees_ids = [m.get('id') for m in self.scope.members()
                                        if m.get('id') in assignees or m.get('username') in set(assignees)]

                if len(update_assignees_ids) != len(assignees):
                    raise NotFoundError('All assignees should be a member of the project.')
            else:
                update_assignees_ids = list()

            update_dict['assignees_ids'] = update_assignees_ids

        if kwargs:
            update_dict.update(kwargs)

        return update_dict

    def delete(self) -> bool:
        """Delete this activity.

        :return: True when successful
        :raises APIError: when unable to delete the activity
        """
        response = self._client._request('DELETE', self._client._build_url('activity', activity_id=self.id))

        if response.status_code != requests.codes.no_content:
            raise APIError("Could not delete Activity {}.".format(self), response=response)
        return True

    #
    # Searchers and retrievers
    #

    def parts(self, *args, **kwargs):
        """Retrieve parts belonging to this activity.

        Without any arguments it retrieves the Instances related to this task only.

        This call only returns the configured properties in an activity. So properties that are not configured
        are not in the returned parts.

        See :class:`pykechain.Client.parts` for additional available parameters.

        Example
        -------
        >>> task = project.activity('Specify Wheel Diameter')
        >>> parts = task.parts()

        To retrieve the models only.
        >>> parts = task.parts(category=Category.MODEL)

        """
        return [p for w in self.widgets() for p in w.parts(*args, **kwargs)]

    def associated_parts(self, *args, **kwargs):
        """Retrieve models and instances belonging to this activity.

        This is a convenience method for the :func:`Activity.parts()` method, which is used to retrieve both the
        `Category.MODEL` as well as the `Category.INSTANCE` in a tuple.

        This call only returns the configured properties in an activity. So properties that are not configured
        are not in the returned parts.

        If you want to retrieve only the models associated to this task it is better to use:
            `task.parts(category=Category.MODEL)`.

        See :func:`pykechain.Client.parts` for additional available parameters.

        :returns: a tuple(models of :class:`PartSet`, instances of :class:`PartSet`)

        Example
        -------
        >>> task = project.activity('Specify Wheel Diameter')
        >>> all_models, all_instances = task.associated_parts()

        """
        associated_models = list()
        associated_instances = list()
        for widget in self.widgets():
            associated_models.extend(widget.parts(category=Category.MODEL, *args, **kwargs))
            associated_instances.extend(widget.parts(category=Category.INSTANCE, *args, **kwargs))

        return (
            associated_models,
            associated_instances
        )

    def associated_object_ids(self) -> List[Dict]:
        """Retrieve object ids associated to this activity.

        This represents a more in-depth retrieval of objects associated to the activity. Each element in the list
        represents a `Property` of `Category.INSTANCE`. Each element contains the following fields:

        'id': The ID of the association
        'widget': The ID of the widget to which the Property instance is associated
        'activity': The ID of the activity
        'model_property': The ID of the Property model
        'model_part': The ID of the model of the Part containing said Property
        'instance_property': The ID of the Property instance
        'instance_part': The ID of the Part instance containing said Property
        'writable': True if the Property is writable, False if is not

        See :func:`pykechain.Client.parts` for additional available parameters.

        :returns: a list of dictonaries with association objects associated to the activity
        :raises NotFoundError: When the response from the server was invalid.

        Example
        -------
        >>> task = project.activity('Specify Wheel Diameter')
        >>> associated_object_ids = task.associated_object_ids()

        """
        request_params = dict(
            activity=self.id,
        )

        url = self._client._build_url('associations')

        response = self._client._request('GET', url, params=request_params)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise NotFoundError("Could not retrieve Associations on Activity {}".format(self), response=response)

        data = response.json()
        return data['results']

    #
    # Customizations
    #

    def widgets(self, **kwargs) -> 'WidgetsManager':
        """
        Widgets of the activity.

        Works with KE-chain version 3.

        :param kwargs: additional keyword arguments
        :return: A :class:`WidgetManager` list, containing the widgets
        :rtype: WidgetManager
        :raises NotFoundError: when the widgets could not be found
        :raises APIError: when the API does not support the widgets, or when the API gives an error.
        """
        if self._widgets_manager is None:
            widgets = self._client.widgets(activity=self.id, **kwargs)
            self._widgets_manager = WidgetsManager(widgets=widgets, activity=self)
        return self._widgets_manager

    def download_as_pdf(
            self,
            target_dir: str = None,
            pdf_filename: str = None,
            paper_size: PaperSize = PaperSize.A4,
            paper_orientation: PaperOrientation = PaperOrientation.PORTRAIT,
            include_appendices: bool = False,
            include_qr_code: bool = False,
    ):
        """
        Retrieve the PDF of the Activity.

        .. versionadded:: 2.1

        :param target_dir: (optional) directory path name where the store the log.txt to.
        :type target_dir: basestring or None
        :param pdf_filename: (optional) log filename to write the log to, defaults to `log.txt`.
        :type pdf_filename: basestring or None
        :param paper_size: The size of the paper to which the PDF is downloaded:
                               - a4paper (default): A4 paper size
                               - a3paper: A3 paper size
                               - a2paper: A2 paper size
                               - a1paper: A1 paper size
                               - a0paper: A0 paper size
        :type paper_size: basestring (see :class:`enums.PaperSize`)
        :param paper_orientation: The orientation of the paper to which the PDF is downloaded:
                               - portrait (default): portrait orientation
                               - landscape: landscape orientation
        :type paper_size: basestring (see :class:`enums.PaperOrientation`)
        :param include_appendices: True if the PDF should contain appendices, False (default) if otherwise.
        :type include_appendices: bool
        :param include_qr_code: True if the PDF should include a QR-code, False (default) if otherwise.
        :type include_qr_code: bool
        :raises APIError: if the pdf file could not be found.
        :raises OSError: if the file could not be written.
        """
        if not pdf_filename:
            pdf_filename = self.name + '.pdf'
        if not pdf_filename.endswith('.pdf'):
            pdf_filename += '.pdf'

        full_path = os.path.join(target_dir or os.getcwd(), pdf_filename)

        request_params = dict(
            papersize=check_enum(paper_size, PaperSize, "paper_size"),
            orientation=check_enum(paper_orientation, PaperOrientation, "paper_orientation"),
            appendices=check_type(include_appendices, bool, "include_appendices"),
            includeqr=check_type(include_qr_code, bool, "include_qr_code"),
        )

        url = self._client._build_url('activity_export', activity_id=self.id)
        response = self._client._request('GET', url, params=request_params)
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not download PDF of Activity {}".format(self), response=response)

        # If appendices are included, the request becomes asynchronous

        if include_appendices:  # pragma: no cover
            data = response.json()

            # Download the pdf async
            url = urljoin(self._client.api_root, data['download_url'])

            count = 0

            while count <= ASYNC_TIMEOUT_LIMIT:
                response = self._client._request('GET', url=url)

                if response.status_code == requests.codes.ok:  # pragma: no cover
                    with open(full_path, 'wb') as f:
                        for chunk in response.iter_content(1024):
                            f.write(chunk)
                    return

                count += ASYNC_REFRESH_INTERVAL
                time.sleep(ASYNC_REFRESH_INTERVAL)

            raise APIError("Could not download PDF of Activity {} within the time-out limit of {} "
                           "seconds".format(self, ASYNC_TIMEOUT_LIMIT), response=response)

        with open(full_path, 'wb') as f:
            for chunk in response.iter_content(1024):
                f.write(chunk)

    def move(self, parent, classification=None):
        """
        Move the `Activity` to a new parent.

        See :func:`pykechain.Client.move_activity` for available parameters.

        If you want to move an Activity from one classification to another, you need to provide the target
        classification. The classification of the parent should match the one provided in the function. This is
        to ensure that you really want this to happen.

        :param parent: parent object to move activity to
        :type parent: Union[Activity, Text]
        :param classification: (optional) classification of the target parent if you want to change the classification.
        :type classification: ActivityClassification or None
        :raises IllegalArgumentError: if the 'parent' activity_type is not :class:`enums.ActivityType.SUBPROCESS`
        :raises IllegalArgumentError: if the 'parent' type is not :class:`Activity` or UUID
        :raises APIError: if an Error occurs.
        """
        return self._client.move_activity(self, parent, classification=classification)

    def share_link(self, subject: Text, message: Text, recipient_users: List[Union[User, Text]]) -> None:
        """
        Share the link of the `Activity` through email.

        :param subject: subject of email
        :type subject: basestring
        :param message: message of email
        :type message: basestring
        :param recipient_users: users that will receive the email
        :type recipient_users: list(Union(User, Id))
        :raises APIError: if an internal server error occurred.
        """
        params = dict(
            message=check_text(message, 'message'),
            subject=check_text(subject, 'subject'),
            recipient_users=[check_user(recipient, User, 'recipient') for recipient in recipient_users],
            activity_id=self.id
        )

        url = self._client._build_url('notification_share_activity_link')

        response = self._client._request('POST', url, data=params)

        if response.status_code not in (requests.codes.created, requests.codes.accepted):  # pragma: no cover
            raise APIError("Could not share the link to Activity {}".format(self), response=response)

    def share_pdf(
            self,
            subject: Text,
            message: Text,
            recipient_users: List[Union[User, Text]],
            paper_size: Optional[PaperSize] = PaperSize.A3,
            paper_orientation: Optional[PaperOrientation] = PaperOrientation.PORTRAIT,
            include_appendices: Optional[bool] = False,
            include_qr_code: Optional[bool] = False,
    ) -> None:
        """
        Share the PDF of the `Activity` through email.

        :param subject: subject of email
        :type subject: basestring
        :param message: message of email
        :type message: basestring
        :param recipient_users: users that will receive the email
        :type recipient_users: list(Union(User, Id))
        :param paper_size: The size of the paper to which the PDF is downloaded:
                               - a4paper: A4 paper size
                               - a3paper: A3 paper size (default)
                               - a2paper: A2 paper size
                               - a1paper: A1 paper size
                               - a0paper: A0 paper size
        :type paper_size: basestring (see :class:`enums.PaperSize`)
        :param paper_orientation: The orientation of the paper to which the PDF is downloaded:
                               - portrait (default): portrait orientation
                               - landscape: landscape orientation
        :type paper_size: basestring (see :class:`enums.PaperOrientation`)
        :param include_appendices: True if the PDF should contain appendices, False (default) if otherwise.
        :type include_appendices: bool
        :param include_qr_code: True if the PDF should include a QR-code, False (default) if otherwise.
        :type include_qr_code: bool
        :raises APIError: if an internal server error occurred.
        """
        recipient_emails = list()
        recipient_users_ids = list()
        if isinstance(recipient_users, list) and all(isinstance(r, (str, int, User)) for r in recipient_users):
            for user in recipient_users:
                if is_valid_email(user):
                    recipient_emails.append(user)
                else:
                    recipient_users_ids.append(check_user(user, User, 'recipient'))
        else:
            raise IllegalArgumentError('`recipients` must be a list of User objects, IDs or email addresses, '
                                       '"{}" ({}) is not.'.format(recipient_users, type(recipient_users)))

        params = dict(
            message=check_text(message, "message"),
            subject=check_text(subject, "subject"),
            recipient_users=recipient_users_ids,
            recipient_emails=recipient_emails,
            activity_id=self.id,
            papersize=check_enum(paper_size, PaperSize, "paper_size"),
            orientation=check_enum(paper_orientation, PaperOrientation, "paper_orientation"),
            appendices=check_type(include_appendices, bool, "include_appendices"),
            includeqr=check_type(include_qr_code, bool, "include_qr_code"),
        )

        url = self._client._build_url('notification_share_activity_pdf')

        response = self._client._request('POST', url, data=params)

        if response.status_code not in (requests.codes.created, requests.codes.accepted):  # pragma: no cover
            raise APIError("Could not share the link to Activity {}".format(self), response=response)



from typing import Iterable, Text  # noqa: F401

from pykechain.models.part import Part  # noqa: F401


class PartSet(Iterable):
    """A set of KE-chain parts.

    Adding set-like methods on a list of parts:
     * iterable
     * len()
     * get()
     * iPython notebook support for HTML table
    """

    def __init__(self, parts: Iterable[Part]):
        """Construct a PartSet from a part iterable."""
        self._parts = list(parts)

    def __repr__(self):  # pragma: no cover
        return "<pyke {} object {} parts>".format(self.__class__.__name__, self.__len__())

    def __iter__(self):
        return iter(self._parts)

    def __len__(self):
        return len(self._parts)

    def __getitem__(self, k: int) -> Part:
        if isinstance(k, int):
            return self._parts[k]

        raise NotImplementedError

    def _repr_html_(self) -> Text:
        all_instances = all(p.category == 'INSTANCE' for p in self._parts)

        html = [
            "<table width=100%>",
            "<tr>",
            "<th>Part</th>"
        ]

        if not all_instances:
            html.append("<th>Category</th>")

        html.append("<th>ID</th>")
        html.append("</tr>")

        for part in self._parts:
            html.append("<tr>")
            html.append("<td>{}</td>".format(part.name))

            if not all_instances:
                html.append("<td>{}</td>".format(part.category))

            html.append("<td>{}</td>".format(part.id))
            html.append("</tr>")

        html.append("</table>")

        return ''.join(html)



from pykechain.exceptions import _DeprecationMixin


class ActivityReferencesProperty(_DeprecationMixin):
    """A virtual object representing a KE-chain Activity References property.

    .. versionadded:: 3.7
    """

    pass



import datetime
from unittest import skipIf

import pytest

from pykechain.enums import Multiplicity, Category, Classification, PropertyType
from pykechain.exceptions import NotFoundError, MultipleFoundError, APIError, IllegalArgumentError
from pykechain.models import PartSet, Part
from tests.classes import TestBetamax
from tests.utils import TEST_FLAG_IS_PIM3


class TestParts(TestBetamax):

    def setUp(self):
        super().setUp()
        self._part = None

    def tearDown(self):
        if self._part:
            self._part.delete()
        super().tearDown()

    def test_retrieve_parts(self):
        parts = self.project.parts()

        # Check if there are parts
        self.assertTrue(len(parts))
        self.assertIsInstance(parts[0], Part)

    def test_retrieve_single_part(self):
        part_to_retrieve = self.project.parts()[0]
        part = self.project.part(part_to_retrieve.name)

        self.assertTrue(part)

    def test_base_comparison(self):
        all_parts = self.project.parts()

        tip = ' See Base.__eq__() method.'
        self.assertFalse(all_parts[0] == all_parts[1], msg='2 different parts must not be equal.' + tip)
        self.assertTrue(all_parts[0] == all_parts[0], msg='The same part must be equal.' + tip)
        self.assertFalse(all_parts[0] == 5, msg='A part must not be equal to a non-pykechain object.' + tip)

    def test_base_hash(self):
        wheel = self.project.model('Wheel')
        a_dict = dict()

        try:
            a_dict[wheel] = 3
        except TypeError:
            self.assertTrue(False, msg='Parts must be hashable See Base.__hash__() method.')

    def test_part_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'name', 'created_at', 'updated_at', 'ref',
                      'category', 'classification', 'parent_id', 'description', 'multiplicity',
                      '_cached_children', 'properties']

        obj = self.project.parts(limit=1)[0]
        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(hasattr(obj, attribute), "Could not find '{}' in the object: '{}'".format(
                    attribute, obj.__dict__.keys()))

    def test_retrieve_single_unknown(self):
        with self.assertRaises(NotFoundError):
            self.project.part('123lladadwd')

    def test_retrieve_single_multiple(self):
        with self.assertRaises(MultipleFoundError):
            self.project.part()

    def test_retrieve_models(self):
        wheel = self.project.model('Wheel')

        self.assertTrue(self.project.parts(model=wheel))

    def test_retrieve_model_unknown(self):
        with self.assertRaises(NotFoundError):
            self.client.model('123lladadwd')

    def test_retrieve_model_multiple(self):
        with self.assertRaises(MultipleFoundError):
            self.project.model()

    def test_scope(self):
        first_part = self.project.parts()[0]

        self.assertEqual(first_part.scope(), self.project)

    def test_part_set_iterator(self):
        for part in self.project.parts():
            self.assertTrue(part.name)

    def test_part_set_get_item_invalid(self):
        part_set = self.project.parts()

        with self.assertRaises(NotImplementedError):
            # noinspection PyStatementEffect
            part_set['testing']

    def test_wrongly_create_model(self):
        # setUp
        bike_model = self.project.model(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self.client.create_model(name='Multiplicity does not exist', parent=bike_model, multiplicity='TEN')

    def test_part_add_delete_part(self):
        bike = self.project.part('Bike')
        wheel_model = self.project.model('Wheel')

        wheel = bike.add(wheel_model, name='Test Wheel')
        wheel.delete()

        wheel = wheel_model.add_to(bike)
        wheel.delete()

        with self.assertRaises(APIError):
            bike.delete()

    def test_create_part_where_parent_is_model(self):
        # setUp
        bike_model = self.project.model(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self._part = self.client.create_part(name='Parent should be instance',
                                                 parent=bike_model, model=bike_model)

    def test_create_part_where_model_is_instance(self):
        # setUp
        bike_instance = self.project.part(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self._part = self.client.create_part(name='Model should not be instance',
                                                 parent=bike_instance, model=bike_instance)

    def test_create_model_where_parent_is_instance(self):
        # setUp
        bike_instance = self.project.part(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self._part = self.client.create_model(name='Parent should be model',
                                                  parent=bike_instance, multiplicity=Multiplicity.ONE)

    def test_create_proxy_model_where_model_is_instance(self):
        # setUp
        bike_instance = self.project.part(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self._part = self.client.create_proxy_model(name='Model should not be instance',
                                                        model=bike_instance, parent=bike_instance)

    def test_create_proxy_model_where_parent_is_instance(self):
        # setUp
        bike_instance = self.project.part(name='Bike')
        bike_model = self.project.model(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self._part = self.client.create_proxy_model(name='Parent should not be instance',
                                                        model=bike_model, parent=bike_instance)

    def test_add_to_wrong_categories(self):
        # This test has the purpose of testing of whether APIErrors are raised when illegal operations are
        # performed (e.g. operation that should be performed on part instances but instead are being performed
        # on part models and vice versa)
        project = self.project

        bike_model = project.model('Bike')
        bike_instance = project.part('Bike')
        wheel_model = project.model('Wheel')

        with self.assertRaises(APIError):
            self._part = bike_model.add(wheel_model)

        front_fork_instance = project.part('Front Fork')

        with self.assertRaises(APIError):
            self._part = front_fork_instance.add_to(bike_instance)

        with self.assertRaises(APIError):
            self._part = bike_instance.add_model('Engine')

        with self.assertRaises(APIError):
            self._part = bike_instance.add_property('Electrical Power')

        with self.assertRaises(APIError):
            self._part = bike_model.add_with_properties(wheel_model)

    def test_part_html_table(self):
        part = self.project.part('Front Wheel')

        self.assertIn("</table>", part._repr_html_())

    def test_part_set_html_table(self):
        parts = self.project.parts()

        self.assertIn("</table>", parts._repr_html_())

    def test_part_set_html_categories(self):
        parts = self.project.parts(category=None)

        self.assertIn("<th>Category</th>", parts._repr_html_())

    # version 1.1.2 and later
    def test_part_set_with_limit(self):
        limit = 5
        parts = self.project.parts(limit=limit)

        self.assertEqual(len(parts), limit)

    def test_part_set_with_batch(self):
        batch = 5
        parts = self.project.parts(batch=batch)
        self.assertTrue(len(parts) >= batch)

    # version 1.1.3 and later
    def test_retrieve_parent_of_part(self):
        frame = self.project.part('Frame')  # type: Part
        self.assertTrue(hasattr(frame, 'parent_id'))
        parent_of_frame = frame.parent()
        self.assertIsInstance(parent_of_frame, type(frame))

    def test_retrieve_children_of_part(self):
        bike = self.project.part('Bike')  # type: Part
        self.assertIsInstance(bike, Part)
        children_of_bike = bike.children()
        self.assertIsInstance(children_of_bike, (PartSet, list))
        self.assertTrue(len(children_of_bike) >= 1)  # eg. Wheels ...

    def test_retrieve_siblings_of_part(self):
        """Test if the siblings of a part is the same as the children of the parent of the part"""
        frame = self.project.part('Frame')  # type: Part
        siblings_of_frame = frame.siblings()
        self.assertIsInstance(siblings_of_frame, PartSet)
        self.assertTrue(len(siblings_of_frame) >= 1)  # eg. Wheels ...

        # double check that the children of the parent of frame are the same as the siblings of frame
        children_of_parent_of_frame = frame.parent().children()  # type: PartSet
        self.assertEqual(len(children_of_parent_of_frame), len(siblings_of_frame))
        children_of_parent_of_frame_ids = [p.id for p in children_of_parent_of_frame]
        for sibling in siblings_of_frame:
            self.assertIn(sibling.id, children_of_parent_of_frame_ids,
                          'sibling {} is appearing in the siblings method and not in the children of '
                          'parent method'.format(sibling))

    # new in 1.3+
    def test_kwargs_on_part_retrieval(self):
        # test that the additional kwargs are added to the query filter on the api
        bikes = self.project.parts('Bike', order=True)  # type:PartSet
        self.assertTrue(len(bikes) == 1)
        self.assertTrue(self.client.last_url.find('order'))

    # new in 1.5+
    def test_edit_part_instance_name(self):
        front_fork = self.project.part('Front Fork')
        front_fork_old_name = str(front_fork.name)
        front_fork.edit(name='Front Fork - updated')

        front_fork_updated = self.project.part('Front Fork - updated')
        self.assertEqual(front_fork.id, front_fork_updated.id)
        self.assertEqual(front_fork.name, front_fork_updated.name)
        self.assertEqual(front_fork.name, 'Front Fork - updated')

        with self.assertRaises(IllegalArgumentError):
            front_fork.edit(name=True)

        # teardown
        front_fork.edit(name=front_fork_old_name)

    def test_edit_part_instance_description(self):
        front_fork = self.project.part('Front Fork')
        front_fork_old_description = str(front_fork._json_data.get('description'))
        front_fork.edit(description='A normal Front Fork')

        front_fork_updated = self.project.part('Front Fork')
        self.assertEqual(front_fork.id, front_fork_updated.id)

        with self.assertRaises(IllegalArgumentError):
            front_fork.edit(description=42)

        # teardown
        front_fork.edit(description=front_fork_old_description)

    def test_edit_part_model_name(self):
        front_fork = self.project.model('Front Fork')
        front_fork_oldname = str(front_fork.name)
        front_fork.edit(name='Front Fork - updated')

        front_fork_updated = self.project.model('Front Fork - updated')
        self.assertEqual(front_fork.id, front_fork_updated.id)
        self.assertEqual(front_fork.name, front_fork_updated.name)

        # teardown
        front_fork.edit(name=front_fork_oldname)

    # test added due to #847 - providing no inputs overwrites values
    def test_edit_part_clear_values(self):
        # setup
        front_fork = self.project.part('Front Fork')
        front_fork_name = str(front_fork.name)
        front_fork_description = str(front_fork.description)

        front_fork.edit(
            name=None,
            description=None
        )

        # testing
        self.assertEqual(front_fork.name, front_fork_name)
        self.assertEqual(front_fork.description, str())

        # tearDown
        front_fork.edit(description=front_fork_description)

    def test_create_model(self):
        bike = self.project.model('Bike')
        pedal = self.project.create_model(bike, 'Pedal', multiplicity=Multiplicity.ONE)
        self._part = pedal
        pedal_model = self.project.model('Pedal')

        self.assertTrue(pedal.id, pedal_model.id)
        self.assertTrue(pedal._json_data['multiplicity'], 'ONE')

    def test_add_proxy_to(self):
        catalog_container = self.project.model(name__startswith='Catalog')
        bearing_catalog_model = catalog_container.add_model('Bearing', multiplicity=Multiplicity.ZERO_MANY)
        wheel_model = self.project.model('Wheel')

        bearing_proxy_model = bearing_catalog_model.add_proxy_to(wheel_model, 'Bearing', Multiplicity.ZERO_ONE)

        self.assertTrue(bearing_proxy_model.category, Category.MODEL)
        self.assertTrue(bearing_proxy_model.parent(), wheel_model)

        # teardown
        all_bearing_proxies = self.project.parts(name='Bearing', category=Category.MODEL, parent=wheel_model.id)
        self.assertGreaterEqual(len(all_bearing_proxies), 1)
        for bearing_proxy in all_bearing_proxies:
            bearing_proxy.delete()

        all_bearing_catalog_models = self.project.parts(name='Bearing', category=Category.MODEL)
        self.assertGreaterEqual(len(all_bearing_catalog_models), 1)
        for bearing_catalog_model in all_bearing_catalog_models:
            bearing_catalog_model.delete()

        all_bearing_models = self.project.parts(name='Bearing', category=Category.MODEL)
        self.assertEqual(len(all_bearing_models), 0)

    # new in 1.6
    def test_retrieve_model(self):
        front_fork = self.project.part('Front Fork')
        front_fork_model = self.project.model('Front Fork')

        front_fork_retrieved_model = front_fork.model()

        # Added to improve coverage. Assert whether NotFoundError is raised when model() method is applied to
        # a part that has no model
        with self.assertRaises(NotFoundError):
            front_fork_model.model()

        self.assertEqual(front_fork_model.id, front_fork_retrieved_model.id)

    def test_count_instances(self):
        front_fork_model = self.project.model("Wheel")
        nr = front_fork_model.count_instances()

        self.assertIsInstance(nr, int)
        self.assertEqual(2, nr)

        bike_part = self.project.part("Bike")
        with self.assertRaises(IllegalArgumentError):
            bike_part.count_instances()

    def test_count_children(self):
        bike_instance = self.project.part("Bike")
        bike_model = self.project.model("Bike")

        nr = bike_instance.count_children()

        self.assertIsInstance(nr, int)
        self.assertEqual(7, nr)

        nr = bike_model.count_children()

        self.assertEqual(5, nr)

    def test_retrieve_catalog_model_of_proxy(self):
        catalog_container = self.project.model(name__startswith='Catalog')
        bearing_catalog_model = catalog_container.add_model('Bearing', multiplicity=Multiplicity.ZERO_MANY)
        wheel_model = self.project.model('Wheel')

        # add proxy model to product Bike > Wheel based on catalog model 'Bearing'
        bearing_proxy_model = bearing_catalog_model.add_proxy_to(wheel_model, 'Bearing', Multiplicity.ZERO_ONE)

        self.assertTrue(bearing_proxy_model.category, Category.MODEL)
        self.assertTrue(bearing_proxy_model.parent(), wheel_model)

        # the call to test here
        re_retrieved_bearing_catalog_model = self.project.model('Bearing', classification=Classification.CATALOG)
        self.assertEqual(bearing_catalog_model, re_retrieved_bearing_catalog_model)

        # teardown
        all_bearing_proxies = self.project.parts(name='Bearing', category=Category.MODEL, parent=wheel_model.id)
        self.assertGreaterEqual(len(all_bearing_proxies), 1)
        for bearing_proxy in all_bearing_proxies:
            bearing_proxy.delete()

        all_bearing_catalog_models = self.project.parts(name='Bearing', category=Category.MODEL)
        self.assertGreaterEqual(len(all_bearing_catalog_models), 1)
        for bearing_catalog_model in all_bearing_catalog_models:
            bearing_catalog_model.delete()

        all_bearing_models = self.project.parts(name='Bearing', category=Category.MODEL)
        self.assertEqual(len(all_bearing_models), 0)

    def test_retrieve_non_existent_proxies_of_a_catalog_model_raises_error(self):
        # Added to improve coverage. Assert whether NotFoundError is raised when proxy_model() method is applied to
        # a part that is not a proxy
        catalog_model = self.project.model('Model', classification=Classification.CATALOG)
        with self.assertRaises(NotFoundError):
            catalog_model.proxy_model()

    def test_retrieve_proxy_of_instance(self):
        instance = self.project.part('Rear Wheel')

        with self.assertRaises(IllegalArgumentError):
            instance.proxy_model()

    # new in 1.8+
    def test_retrieve_part_multiplicity(self):
        bike_model = self.project.model('Bike')
        self.assertEqual(bike_model.multiplicity, Multiplicity.ONE_MANY, bike_model.multiplicity)

    # new in 1.9
    def test_retrieve_part_properties_in_a_dict(self):
        # Retrieve the bike part
        bike = self.project.part('Bike')

        # Call the function to be tested
        bike_properties = bike.as_dict()

        # Check whether bike_properties contains all the property names in bike
        for prop in bike.properties:
            self.assertTrue(prop.name in bike_properties)

    # new in 1.12
    def test_retrieve_children_of_part_with_additional_arguments(self):
        bike = self.project.part('Bike')  # type: Part
        self.assertIsInstance(bike, Part)
        children_of_bike = bike.children(name__icontains='Wheel')
        self.assertIsInstance(children_of_bike, PartSet)
        self.assertTrue(len(children_of_bike) >= 1)  # eg. Wheels ...

    def test_retrieve_siblings_of_part_with_additional_arguments(self):
        """Test if the siblings of a part is the same as the children of the parent of the part"""
        frame = self.project.part('Frame')  # type: Part
        siblings_of_frame = frame.siblings(name__icontains='Wheel')
        self.assertIsInstance(siblings_of_frame, PartSet)
        self.assertTrue(len(siblings_of_frame) >= 1)  # eg. Wheels ...

    # new in 2.3
    def test_clone_model(self):
        # setUp
        model_name = 'Seat'
        seat = self.project.model(model_name)
        self._part = seat.clone()

        # testing
        clone_seat_model = self.project.model('CLONE - {}'.format(model_name))
        self.assertTrue(clone_seat_model)

    def test_clone_instance(self):
        instance_name = 'Front Wheel'
        wheel = self.project.part(instance_name)
        self._part = wheel.clone()

        # testing
        clone_spoke_instance = self.project.part('CLONE - {}'.format(instance_name))
        self.assertTrue(clone_spoke_instance)

    def test_clone_instance_with_multiplicity_violation(self):
        instance_name = 'Seat'
        seat = self.project.part(instance_name)

        seat_model = seat.model()

        # testing
        self.assertEqual(Multiplicity.ONE, seat_model.multiplicity)
        with self.assertRaises(APIError):
            self._part = seat.clone()


@skipIf(not TEST_FLAG_IS_PIM3, reason="This tests is designed for PIM version 3, expected to fail on older PIM3")
class TestPIM2SpecificPartTests(TestBetamax):
    """Pim3 capable tests."""

    def test_retrieve_part_without_parent_id(self):
        # only the root does not have a parent_id
        product_root_node = self.project.part(name='Product', classification=Classification.PRODUCT)
        self.assertTrue(hasattr(product_root_node, 'parent_id'))
        self.assertIsNotNone(product_root_node.parent_id)
        self.assertIsNone(product_root_node.parent().parent_id)

    def test_retrieve_parent_of_part_without_parent_id(self):
        # only the root does not have a parent_id
        product_root_node = self.project.part(name='Product', classification=Classification.PRODUCT)
        root_node = product_root_node.parent()
        self.assertEqual(root_node.name, "Root")
        self.assertIsNone(root_node.parent())

    def test_retrieve_siblings_of_part_without_parent_id(self):
        product_root_node = self.project.part(name='Product', classification=Classification.PRODUCT)
        siblings_of_product_root_node = product_root_node.siblings()
        self.assertIsInstance(siblings_of_product_root_node, PartSet)
        self.assertEqual(len(siblings_of_product_root_node), 2)

        siblings_of_root_node = product_root_node.parent().siblings()
        self.assertIsInstance(siblings_of_root_node, PartSet)
        self.assertEqual(len(siblings_of_root_node), 0)


@skipIf(not TEST_FLAG_IS_PIM3, reason="This tests is designed for PIM version 3, expected to fail on older PIM3")
@pytest.mark.skipif("os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
                    reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided")
class TestBulkPartsCreation(TestBetamax):
    def setUp(self):
        super().setUp()

        self.product_model = self.project.model(name="Product")
        self.part_model = self.project.create_model(name="Part", parent=self.product_model)
        self.text_prop = self.part_model.add_property(name="Text Property", property_type=PropertyType.TEXT_VALUE, unit="mm")
        self.char_prop = self.part_model.add_property(name="Char Property", property_type=PropertyType.CHAR_VALUE)
        self.int_prop = self.part_model.add_property(name="Int Property", property_type=PropertyType.INT_VALUE)
        self.float_prop = self.part_model.add_property(name="Float Property", property_type=PropertyType.FLOAT_VALUE)
        self.bool_prop = self.part_model.add_property(
            name="Boolean Property", property_type=PropertyType.BOOLEAN_VALUE)
        self.date_prop = self.part_model.add_property(name="Date Property", property_type=PropertyType.DATE_VALUE)
        self.datetime_prop = self.part_model.add_property(
            name="Datetime Property", property_type=PropertyType.DATETIME_VALUE)
        self.attachment_prop = self.part_model.add_property(
            name="Attach Property", property_type=PropertyType.ATTACHMENT_VALUE)
        self.link_prop = self.part_model.add_property(name="Link Property", property_type=PropertyType.LINK_VALUE)
        self.ss_prop = self.part_model.add_property(
            name="SS Property", property_type=PropertyType.SINGLE_SELECT_VALUE)
        self.ss_prop.options = ["1", "2", "3", "4", "5"]
        self.ms_prop = self.part_model.add_property(
            name="MS Property", property_type=PropertyType.MULTI_SELECT_VALUE,
            options={"value_choices": ["1", "2", "3", "4", "5"]})
        self.part_ref_prop = self.part_model.add_property(
            name="Part ref Property", property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.project.model(name="Wheel").id)
        self.geo_info_prop = self.part_model.add_property(
            name="Geo Property", property_type=PropertyType.GEOJSON_VALUE)
        self.weather_prop = self.part_model.add_property(
            name="Weather Property", property_type=PropertyType.WEATHER_VALUE)
        self.act_ref_prop = self.part_model.add_property(
            name="Act Property", property_type=PropertyType.ACTIVITY_REFERENCES_VALUE)

        self.product_part = self.project.part(name="Product")
        self.part_dict = {
            "name": "Dummy name",
            "parent_id": self.product_part.id,
            "description": "",
            "model_id": self.part_model.id,
            "properties": [
                {
                    "name": self.text_prop.name,
                    "value": "Dummy value",
                    "model_id": self.text_prop.id,
                }]
        }

    def tearDown(self):
        self.part_model.delete()

    """Bulk parts creation"""
    def test_bulk_create_parts(self):
        new_parts = list()
        for idx in range(1, 5):
            part_dict = {
                "name": idx,
                "parent_id": self.product_part.id,
                "description": "Description part {}".format(idx),
                "model_id": self.part_model.id,
                "properties": [
                    {
                        "name": self.text_prop.name,
                        "value": "{}".format(idx),
                        "model_id": self.text_prop.id,
                    },
                    {
                        "name": self.char_prop.name,
                        "value": "{}".format(idx),
                        "model_id": self.char_prop.id
                    },
                    {
                        "name": self.int_prop.name,
                        "value": idx,
                        "model_id": self.int_prop.id
                    },
                    {
                        "name": self.float_prop.name,
                        "value": float("{}.{}".format(idx, idx)),
                        "model_id": self.float_prop.id
                    },
                    {
                        "name": self.bool_prop.name,
                        "value": True if idx % 2 == 0 else False,
                        "model_id": self.bool_prop.id
                    },
                    {
                        "name": self.date_prop.name,
                        "value": datetime.date(2020, idx, idx).isoformat(),
                        "model_id": self.date_prop.id
                    },
                    {
                        "name": self.datetime_prop.name,
                        "value": datetime.datetime(2020, idx, idx, 15, 00, 00).isoformat("T"),
                        "model_id": self.datetime_prop.id
                    },
                    {
                        "name": self.link_prop.name,
                        "value": "http://{}.com".format(idx),
                        "model_id": self.link_prop.id
                    },
                    {
                        "name": self.ss_prop.name,
                        "value": "{}".format(idx),
                        "model_id": self.ss_prop.id
                    },
                    {
                        "name": self.ms_prop.name,
                        "value": ["{}".format(idx)],
                        "model_id": self.ms_prop.id
                    },
                    {
                        "name": self.part_ref_prop.name,
                        "value": [self.project.part(name="Front Wheel").id],
                        "model_id": self.part_ref_prop.id
                    },
                    {
                        "name": self.act_ref_prop.name,
                        "value": [self.project.activity(name="Specify wheel diameter").id],
                        "model_id": self.act_ref_prop.id
                    },
                ]
            }
            new_parts.append(part_dict)
        parts_created = self.client._create_parts_bulk(
            parts=new_parts
        )

        # testing
        self.assertEqual(len(parts_created), 4)
        for part_created in parts_created:
            idx = int(part_created.name)
            self.assertEqual(part_created.description, "Description part {}".format(idx))
            self.assertEqual(part_created.category, Category.INSTANCE)
            self.assertEqual(len(part_created.properties), 15)

            self.assertEqual(part_created.property(name=self.text_prop.name).value, str(idx))
            self.assertEqual(part_created.property(name=self.char_prop.name).value, str(idx))
            self.assertEqual(part_created.property(name=self.int_prop.name).value, idx)
            self.assertEqual(part_created.property(name=self.float_prop.name).value, float("{}.{}".format(idx, idx)))
            self.assertEqual(part_created.property(name=self.bool_prop.name).value, True if idx % 2 == 0 else False)
            self.assertEqual(part_created.property(name=self.date_prop.name).value,
                             datetime.date(2020, idx, idx).isoformat())
            self.assertEqual(part_created.property(name=self.datetime_prop.name).value.split("+")[0],
                             datetime.datetime(2020, idx, idx, 15, 00, 00).isoformat("T"))
            self.assertEqual(part_created.property(name=self.link_prop.name).value, "http://{}.com".format(idx))
            self.assertEqual(part_created.property(name=self.ss_prop.name).value, "{}".format(idx))
            self.assertEqual(part_created.property(name=self.ms_prop.name).value, ["{}".format(idx)])
            self.assertEqual(part_created.property(name=self.part_ref_prop.name).value[0].name, "Front Wheel")
            self.assertEqual(part_created.property(name=self.act_ref_prop.name).value[0].name, "Specify wheel diameter")
            self.assertEqual(part_created.property(name=self.weather_prop.name).value, dict())
            self.assertEqual(part_created.property(name=self.geo_info_prop.name).value, dict())

    def test_bulk_create_parts_without_name(self):
        self.part_dict.pop("name")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )

    def test_bulk_create_parts_without_parent_id(self):
        self.part_dict.pop("parent_id")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )

    def test_bulk_create_parts_without_model_id(self):
        self.part_dict.pop("model_id")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )

    def test_bulk_create_parts_without_properties(self):
        self.part_dict.pop("properties")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )

    def test_bulk_create_parts_without_property_name(self):
        self.part_dict["properties"][0].pop("name")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )

    def test_bulk_create_parts_without_property_model_id(self):
        self.part_dict["properties"][0].pop("model_id")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )

    def test_bulk_create_parts_without_property_value(self):
        self.part_dict["properties"][0].pop("value")
        with self.assertRaises(IllegalArgumentError):
            self.client._create_parts_bulk(
                parts=[self.part_dict]
            )


@skipIf(not TEST_FLAG_IS_PIM3, reason="This tests is designed for PIM version 3, expected to fail on older PIM3")
@pytest.mark.skipif("os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
                    reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided")
class TestBulkPartsDeletion(TestBetamax):
    def setUp(self):
        super().setUp()
        self.parts = list()
        self.wheel_model = self.project.model(name="Wheel")
        self.bike_instance = self.project.part(name="Bike")

        self.diameter_prop = self.wheel_model.property(name="Diameter")
        self.spokes_prop = self.wheel_model.property(name="Spokes")
        self.rim_material_prop = self.wheel_model.property(name="Rim Material")
        self.tire_thickness_prop = self.wheel_model.property(name="Tire Thickness")

        for idx in range(1, 5):
            part_dict = {
                "name": "Wheel {}".format(idx),
                "parent_id": self.bike_instance.id,
                "description": "",
                "model_id": self.wheel_model.id,
                "properties": [
                    {
                        "name": self.diameter_prop.name,
                        "value": float("{}.{}".format(idx, idx)),
                        "model_id": self.diameter_prop.id,
                    },
                    {
                        "name": self.spokes_prop.name,
                        "value": "{}".format(idx),
                        "model_id": self.spokes_prop.id
                    },
                    {
                        "name": self.rim_material_prop.name,
                        "value": "Material {}".format(idx),
                        "model_id": self.rim_material_prop.id
                    },
                    {
                        "name": self.tire_thickness_prop.name,
                        "value": float("{}.{}".format(idx, idx)),
                        "model_id": self.tire_thickness_prop.id
                    }
                ]
            }
            self.parts.append(part_dict)
        self.parts_created = list(self.client._create_parts_bulk(
            parts=self.parts
        ))

    def tearDown(self):
        for part in self.parts_created:
            # In case the parts have not been deleted and a clean up is required
            try:
                part.delete()
            except APIError:
                pass

    def test_bulk_delete_parts(self):
        input_parts_and_uuids = [self.parts_created[0],
                                 self.parts_created[1],
                                 self.parts_created[2].id,
                                 self.parts_created[3].id]
        self.client._delete_parts_bulk(parts=input_parts_and_uuids)
        for idx in range(1, 5):
            with self.subTest(idx=idx):
                with self.assertRaises(NotFoundError):
                    self.project.part(name="Wheel {}".format(idx))

    def test_bulk_delete_parts_with_wrong_input(self):
        wrong_input = [self.project.activity(name="Specify wheel diameter")]
        with self.assertRaises(IllegalArgumentError):
            self.client._delete_parts_bulk(parts=wrong_input)



from pykechain.exceptions import _DeprecationMixin


class MultiReferenceProperty2(_DeprecationMixin):
    """A virtual object representing a KE-chain multi-references property.

    .. versionadded:: 1.14
    """

    pass



import datetime
from typing import Union, Text
from pykechain.models import Property
from pykechain.models.property2_datetime import DatetimeProperty2
from pykechain.models.input_checks import check_datetime
from pykechain.utils import parse_datetime


class DatetimeProperty(Property, DatetimeProperty2):
    """A virtual object representing a KE-chain datetime property."""

    @property
    def value(self):
        """Retrieve the data value of a property.

        Setting this value will immediately update the property in KE-chain.

        :returns: the value
        """
        return self._value

    @value.setter
    def value(self, value):
        value = check_datetime(dt=value, key='value')

        value = self.serialize_value(value)
        if self.use_bulk_update:
            self._pend_update(dict(value=value))
            self._value = value
        else:
            self._put_value(value)

    def to_datetime(self) -> Union[type(None), datetime.datetime]:
        """Retrieve the datetime as a datetime.datetime value.

        :returns: the value
        """
        return parse_datetime(self._value)

    @staticmethod
    def to_iso_format(date_time: datetime.datetime) -> Text:
        """Convert a datetime object to isoformat."""
        return date_time.isoformat()

    def serialize_value(self, value) -> Text:
        """
        Serialize the value to be set on the property by checking for datetime objects.

        :param value: non-serialized value
        :type value: Any
        :return: serialized value
        """
        if isinstance(value, datetime.datetime):
            value = self.to_iso_format(value)
        return value



import os

from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models import Part
from tests.classes import TestBetamax


class TestPartUpdate(TestBetamax):

    def setUp(self):
        super().setUp()
        self.model = self.project.model("Wheel")
        bike = self.project.part("Bike")
        self.wheel = bike.add(model=self.model, name='_WHEEL')  # type: Part

    def tearDown(self):
        self.wheel.delete()
        super().tearDown()

    def test(self):
        update_dict = {
            'Diameter': 432.1,
            'Spokes': 0,
            'Rim Material': 'Adamantium'
        }
        self.wheel.update(
            name='Better wheel',
            update_dict=update_dict,
        )
        live_wheel = self.project.part(pk=self.wheel.id)
        self.assertEqual(live_wheel.name, 'Better wheel')

        for name, value in update_dict.items():
            self.assertEqual(value, live_wheel.property(name=name).value,
                             "property {} with value {} did not match contents with KEC".format(name, value))

    def test_model(self):
        update_dict = {
            'Diameter': 432.1,
            'Spokes': 0,
            'Rim Material': 'Adamantium'
        }
        self.model.update(update_dict=update_dict)

    def test_with_missing_property(self):
        update_dict = {
            'Unknown Property': 'Woot!'
        }
        with self.assertRaises(NotFoundError):
            self.wheel.update(update_dict=update_dict)

    # noinspection PyTypeChecker
    def test_invalid_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.wheel.update(name=12)

        with self.assertRaises(IllegalArgumentError):
            self.wheel.update(update_dict=("Diameter", 41.2))

        with self.assertRaises(IllegalArgumentError):
            self.wheel.update(properties_fvalues=("Diameter", 41.2))

    # new in 1.14.1
    def test_with_property_ids(self):
        update_dict = dict()
        for p in self.wheel.properties:
            if p.name == 'Diameter':
                update_dict[p.id] = 1.5
            elif p.name == 'Spokes':
                update_dict[p.id] = 20
            elif p.name == 'Rim Material':
                update_dict[p.id] = "Adamantium"

        # do tests
        self.wheel.update(update_dict=update_dict)

    def test_with_attachment(self):
        # setup
        bike_part = self.project.part('Bike')  # type: Part
        bike_part.property(name='Picture').value = None

        project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/'))
        attachment_path = project_root + '/requirements.txt'
        update_dict = {
            'Picture': attachment_path,
        }

        # do tests
        bike_part.update(update_dict=update_dict)
        bike_part.refresh()
        attachment_prop = bike_part.property(name='Picture')

        self.assertTrue(attachment_prop.has_value(), msg='Attachment was not uploaded.')
        self.assertIn('requirements', attachment_prop.filename, msg='filename changed.')

        # tearDown
        attachment_prop.value = None

    def test_bulk_update(self):
        from pykechain.models import Property
        Property.set_bulk_update(True)

        # setUp
        frame = self.project.part("Frame")
        front_wheel = self.project.part("Front Wheel")
        old_data = frame.as_dict()
        update_dict = {
            "Material": "Kevlar",
            "Color": "Black",
            "Ref to wheel": self.wheel,
        }

        # testing
        frame.update(update_dict=update_dict)

        self.assertEqual("Kevlar", frame.property("Material").value)
        self.assertEqual("Black", frame.property("Color").value)
        self.assertEqual(self.wheel.id, frame.property("Ref to wheel").value_ids()[0])
        self.assertEqual(self.wheel, frame.property("Ref to wheel").value[0])

        live_frame = self.project.part("Frame")

        self.assertIsNot(frame, live_frame, msg="Frames must be 2 different Python objects, by memory allocation")
        self.assertEqual(frame, live_frame, msg="Frames must be identical KE-chain objects, by UUID")
        self.assertEqual("Aluminum", live_frame.property("Material").value)
        self.assertEqual("KE-works orange", live_frame.property("Color").value)
        self.assertEqual(front_wheel, live_frame.property("Ref to wheel").value[0])

        Property.update_values(client=self.client)
        live_frame.refresh()

        self.assertEqual("Kevlar", live_frame.property("Material").value)
        self.assertEqual("Black", live_frame.property("Color").value)
        self.assertEqual(self.wheel, live_frame.property("Ref to wheel").value[0])

        # tearDown
        frame.update(update_dict=old_data)


from unittest import TestCase

from pykechain.utils import is_url, is_valid_email, Empty, get_in_chunks
from tests.classes import SixTestCase


class TestIsURL(SixTestCase):
    def test_is_url_returns_true_on_valid_url(self):
        addresses = [
            u"http://foobar.dk",
            u"http://foobar.museum/foobar",
            u"http://fo.com",
            u"http://FOO.com",
            u"http://foo.com/blah_blah",
            u"http://foo.com/blah_blah/",
            u"http://foo.com/blah_blah_(wikipedia)",
            u"http://foo.com/blah_blah_(wikipedia)_(again)",
            u"http://www.example.com/wpstyle/?p=364",
            u"https://www.example.com/foo/?bar=baz&inga=42&quux",
            u"https://www.example.com?bar=baz",
            u"http://✪df.ws/123",
            u"http://userid:password@example.com:8080",
            u"http://userid@example.com",
            u"http://userid@example.com:8080/",
            u"http://userid:password@example.com",
            u"http://142.42.1.1/",
            u"http://142.42.1.1:8080/",
            u"http://⌘.ws/",
            u"http://foo.com/blah_(wikipedia)#cite-1",
            u"http://foo.com/unicode_(✪)_in_parens",
            u"http://foo.com/(something)?after=parens",
            u"http://☺.damowmow.com/",
            u"http://code.google.com/events/#&product=browser",
            u"http://j.mp",
            u"ftp://foo.bar/baz",
            u"http://foo.bar/?q=Test%20URL-encoded%20stuff",
            u"http://-.~_!$&'()*+,;=:%40:80%2f::::::@example.com",
            u"http://1337.net",
            u"http://a.b-c.de",
            u"http://223.255.255.254",
            u"http://127.0.10.150",
            u"http://localhost",
            u"http://localhost:8000",
            u"http://[FEDC:BA98:7654:3210:FEDC:BA98:7654:3210]:80/index.html",
            u"http://[1080:0:0:0:8:800:200C:417A]/index.html",
            u"http://[3ffe:2a00:100:7031::1]",
            u"http://[1080::8:800:200C:417A]/foo",
            u"http://[::192.9.5.5]/ipng",
            u"http://[::FFFF:129.144.52.38]:80/index.html",
            u"http://[2010:836B:4179::836B:4179]",
        ]
        for address in addresses:
            with self.subTest(address):
                self.assertTrue(is_url(address),  "should be a valid address: '{}'".format(address))

    def test_is_url_returns_False_on_failed_url(self):
        failed_addresses = [
            "http://foobar",
            "foobar.dk",
            "http://127.0.0/asdf",
            "http://foobar.d",
            "http://foobar.12",
            "http://foobar",
            "htp://foobar.com",
            "http://foobar..com",
            "http://fo..com",
            "http://",
            "http://.",
            "http://..",
            "http://../",
            "http://?",
            "http://??",
            "http://??/",
            "http://#",
            "http://##",
            "http://##/",
            "http://foo.bar?q=Spaces should be encoded",
            "//",
            "//a",
            "///a",
            "///",
            "http:///a",
            "foo.com",
            "rdar://1234",
            "h://test",
            "http:// shouldfail.com",
            ":// should fail",
            "http://foo.bar/foo(bar)baz quux",
            "ftps://foo.bar/",
            "http://-error-.invalid/",
            "http://a.b--c.de/",
            "http://-a.b.co",
            "http://a.b-.co",
            "http://0.0.0.0",
            "http://10.1.1.0",
            "http://10.1.1.255",
            "http://224.1.1.1",
            "http://1.1.1.1.1",
            "http://123.123.123",
            "http://3628126748",
            "http://.www.foo.bar/",
            "http://www.foo.bar./",
            "http://.www.foo.bar./",
            "http://127.12.0.260",
            'http://example.com/">user@example.com',
            "http://[2010:836B:4179::836B:4179",
            "http://2010:836B:4179::836B:4179",
            "http://2010:836B:4179::836B:4179:80/index.html",
        ]
        for address in failed_addresses:
            with self.subTest(address):
                self.assertFalse(is_url(address), "should be a invalid address: '{}'".format(address))


class TestIsEmail(SixTestCase):
    def test_is_email_returns_true_on_valid_url(self):
        valid_addresses = [
            "some@domain.org",
            "email@example.com",
            "firstname.lastname@example.com",
            "email@subdomain.example.com",
            "firstname+lastname@example.com",
            "email@123.123.123.123",
            "email@[123.123.123.123]",
            '"email"@example.com',
            "1234567890@example.com",
            "email@example-one.com",
            "_______@example.com",
            "email@example.name",
            "email@example.museum",
            "email@example.co.jp",
            "firstname-lastname@example.com",
        ]
        for email in valid_addresses:
            with self.subTest(email):
                self.assertTrue(is_valid_email(email),  "should be a valid address: '{}'".format(email))

    def test_is_email_returns_false_on_invalid_url(self):
        invalid_addresses = [
            "plainaddress",
            "#@%^%#$@#$@#.com",
            "@example.com",
            "Joe Smith <email@example.com>",
            "email.example.com",
            "email@example@example.com",
            ".email@example.com",
            "email.@example.com",
            "email..email@example.com",
            "あいうえお@example.com",
            "email@example.com (Joe Smith)",
            "email@example",
            "email@-example.com",
            "email@example..com",
            "Abc..123@example.com",
            "”(),:;<>[\]@example.com",
            "just”not”right@example.com",
            'this\ is"really"not\allowed@example.com',
        ]
        for email in invalid_addresses:
            with self.subTest(email):
                self.assertFalse(is_valid_email(email), "should be an invalid address: '{}'".format(email))


class TestEmpty(TestCase):

    def test_singleton(self):
        empty_1 = Empty()
        empty_2 = Empty()

        self.assertEqual(empty_1, empty_2)
        self.assertIs(empty_1, empty_2)
        self.assertIsNot(empty_1, Empty)


class TestChunks(TestCase):

    def test_get_in_chunks(self):
        chunks = get_in_chunks(
            lst=list(range(77)),
            chunk_size=9,
        )

        import types
        self.assertIsInstance(chunks, types.GeneratorType)

        chunks_list = list(chunks)
        self.assertEqual(9, len(chunks_list))



import warnings
from requests import Response, PreparedRequest


class APIError(Exception):
    """A general KE-chain API Error occurred.

    A end-user descriptive message is required.

    :ivar response: response object
    :ivar request: request object that precedes the response
    :ivar msg: error message in the response
    :ivar traceback: traceback in the response (from the KE-chain server)
    :ivar detail: details of the error
    """

    def __init__(self, *args, **kwargs):
        """Initialise the `APIError` with `response`, `request`, `msg`, `traceback` and `detail`.

        :param response:
        :param kwargs:
        """
        self.response = kwargs.pop('response', None)

        if hasattr(self.response, 'request'):
            self.request = self.response.request
        else:
            self.request = kwargs.pop('request', None)

        self.msg, self.traceback, self.detail = None, None, None

        if args:
            arg = args[0]
            context = [arg if isinstance(arg, str) else arg.__repr__()]
        else:
            context = ['Error in request to the server.']  # Default message if `APIError()`, without inputs,  is used.

        import json

        if self.response is not None and isinstance(self.response, Response):
            try:
                response_json = self.response.json()
            except json.decoder.JSONDecodeError:
                response_json = None

            if response_json:
                self.msg = response_json.get('msg')
                self.traceback = response_json.get('traceback', 'provided no traceback.')
                self.detail = response_json.get('detail')
                self.results = response_json.get('results')
            else:
                self.traceback = self.response.text
                self.msg = self.traceback.split(r'\n'[-1])
                self.detail = None
                self.results = None

            context.extend([
                'Server {}'.format(self.traceback),
                'Results:\n{}'.format(json.dumps(self.results, indent=4)),
                'Detail: {}'.format(self.detail),
                'Elapsed: {}'.format(self.response.elapsed),
            ])

        if self.request is not None and isinstance(self.request, PreparedRequest):
            context.extend([
                'Request URL: {}'.format(self.request.url),
                'Request method: {}'.format(self.request.method),
            ])
            if self.request.body:
                try:
                    decoded_body = self.request.body.decode("UTF-8")  # Convert byte-string to string
                except AttributeError:
                    decoded_body = self.request.body  # strings (e.g. from testing cassettes) cant be decoded
                try:
                    body = json.loads(decoded_body)  # Convert string to Python object(s)
                except json.decoder.JSONDecodeError:
                    body = decoded_body.split('&')  # parameters in URL
                context.append('Request data:\n{}'.format(json.dumps(body, indent=4)))  # pretty printing of a json

        message = '\n'.join(context)
        new_args = [message]

        super().__init__(*new_args)


class ForbiddenError(APIError):
    """A login is required."""

    pass


class MultipleFoundError(APIError):
    """Multiple objects are found, while a single object is requested."""

    pass


class NotFoundError(APIError):
    """No object is found."""

    pass


class ClientError(APIError):
    """When instantiating the Client an Error occurred."""

    pass


class IllegalArgumentError(ValueError):
    """Illegal arguments where provided."""

    pass


class InspectorComponentError(Exception):
    """Error in the InspectorComponent."""

    pass


class _DeprecationMixin:

    __notified = False

    def __new__(cls, *args, **kwargs):
        if not cls.__notified:
            warnings.warn('`{n}` is a wrapping class for `{nn}`. `{n}` will be deprecated in November 2020.'
                          .format(n=cls.__name__, nn=str(cls.__name__)[:-1]), PendingDeprecationWarning)
            cls.__notified = True

        return super().__new__(cls)



from pykechain.enums import Enum


class DashboardWidgetSourceScopes(Enum):
    """
    Options to link the Dashboard Widget to a source.

    :cvar TAGGED_SCOPES: source is a list of tags
    :cvar CURRENT_SCOPE: source is the current project
    :cvar SELECTED_SCOPES: source is a list of scopes
    :cvar SUBPROCESS: source is an activity
    :cvar OWN_SCOPES: source is list of scopes based on user
    """

    TAGGED_SCOPES = 'taggedScopes'
    CURRENT_SCOPE = 'project'
    SELECTED_SCOPES = 'selectedScopes'
    SUBPROCESS = 'subProcess'
    OWN_SCOPES = 'scopes'


class DashboardWidgetShowTasks(Enum):
    """
    Options to display information about tasks in a Dashboard widget.

    :cvar TOTAL_TASKS: all the tasks
    :cvar CLOSED_TASKS: closed tasks
    :cvar OVERDUE_TASKS: tasks that had their due date in the past
    :cvar UNASSIGNED_TASKS: tasks that have no assignee
    """

    TOTAL_TASKS = "TOTAL_TASKS"
    CLOSED_TASKS = "CLOSED_TASKS"
    OVERDUE_TASKS = "OVERDUE_TASKS"
    UNASSIGNED_TASKS = "UNASSIGNED_TASKS"


class DashboardWidgetShowScopes(Enum):
    """
    Options to display information about scopes in a Dashboard widget.

    :cvar TOTAL_SCOPES: all the projects
    :cvar CLOSED_SCOPES: closed projects
    :cvar OVERDUE_SCOPES: projects that had their due date in the past
    """

    TOTAL_SCOPES = "TOTAL_SCOPES"
    CLOSED_SCOPES = "CLOSED_SCOPES"
    OVERDUE_SCOPES = "OVERDUE_SCOPES"


class AssociatedObjectId(Enum):
    """
    Options to associate object id's to the meta of a widget (e.g. PaginatedGrid).

    :cvar PART_INSTANCE_ID: id of the part instance
    :cvar PARENT_INSTANCE_ID: id of the parent instance
    :cvar PART_MODEL_ID: id of the part model
    :cvar TEAM_ID: id of the team
    :cvar PROPERTY_INSTANCE_ID: id of the property instance
    :cvar SERVICE_ID: id of the service
    :cvar ACTIVITY_ID: id of the activity
    """

    PART_INSTANCE_ID = "partInstanceId"
    PARENT_INSTANCE_ID = "parentInstanceId"
    PART_MODEL_ID = "partModelId"
    TEAM_ID = 'teamId'
    PROPERTY_INSTANCE_ID = "propertyInstanceId"
    SERVICE_ID = "serviceId"
    ACTIVITY_ID = 'activityId'


class MetaWidget(Enum):
    """Options found in the meta of all widgets."""

    # Common
    TITLE = "title"
    SHOW_TITLE_VALUE = "showTitleValue"
    CUSTOM_TITLE = "customTitle"
    SHOW_DESCRIPTION_VALUE = "showDescriptionValue"
    DESCRIPTION_OPTION_NOTHING = "No description"
    DESCRIPTION_OPTION_CUSTOM = "Custom description"
    CUSTOM_DESCRIPTION = "customDescription"
    COLLAPSED = "collapsed"
    COLLAPSIBLE = "collapsible"
    NO_BACKGROUND = "noBackground"
    NO_PADDING = "noPadding"
    IS_DISABLED = "isDisabled"
    IS_MERGED = "isMerged"

    # Prefilters
    PREFILTERS = "prefilters"
    EXCLUDED_PROPERTY_MODELS = "propmodelsExcl"
    PROPERTY_MODELS = "property_models"
    VALUES = "values"
    FILTERS_TYPE = "filters_type"

    # Paginated and Super Grids widgets
    NAME = "name"
    PROPERTY_VALUE_PREFILTER = "property_value"
    SORTED_COLUMN = "sortedColumn"
    SORTED_DIRECTION = "sortDirection"
    SHOW_COLUMNS = 'showColumns'
    SHOW_FILTERS_VALUE = "showCollapseFiltersValue"
    FILTERS_COLLAPSED = "Collapsed"
    FILTERS_EXPANDED = "Expanded"
    COLLAPSE_FILTERS = "collapseFilters"
    CUSTOM_PAGE_SIZE = "customPageSize"
    SHOW_NAME_COLUMN = "showNameColumn"
    SHOW_IMAGES = "showImages"
    VISIBLE_ADD_BUTTON = "addButtonVisible"
    VISIBLE_EDIT_BUTTON = "editButtonVisible"
    VISIBLE_DELETE_BUTTON = "deleteButtonVisible"
    VISIBLE_CLONE_BUTTON = "cloneButtonVisible"
    VISIBLE_DOWNLOAD_BUTTON = "downloadButtonVisible"
    VISIBLE_UPLOAD_BUTTON = "uploadButtonVisible"
    VISIBLE_INCOMPLETE_ROWS = "incompleteRowsVisible"
    EMPHASIZE_ADD_BUTTON = "primaryAddUiValue"
    EMPHASIZE_EDIT_BUTTON = "primaryEditUiValue"
    EMPHASIZE_CLONE_BUTTON = "primaryCloneUiValue"
    EMPHASIZE_DELETE_BUTTON = "primaryDeleteUiValue"

    # ScopeGrid widgets
    VISIBLE_ACTIVE_FILTER = "activeFilterVisible"
    VISIBLE_SEARCH_FILTER = "searchFilterVisible"
    TAGS = "tags"

    # Attachment viewer widgets
    ALIGNMENT = "alignment"
    IMAGE_FIT = "imageFit"
    SHOW_DOWNLOAD_BUTTON = "showDownloadButton"
    SHOW_FULL_SCREEN_IMAGE_BUTTON = "showFullscreenImageButton"

    # Property Grid widget
    SHOW_HEADERS = "showHeaders"
    CUSTOM_HEIGHT = "customHeight"

    # Service Widget and Service Card Widget
    EMPHASIZE_BUTTON = "emphasizeButton"
    SHOW_DOWNLOAD_LOG = "showDownloadLog"
    SHOW_LOG = "showLog"
    BUTTON_TEXT_DEFAULT = "Default"
    BUTTON_NO_TEXT = "No text"
    BUTTON_TEXT_CUSTOM = "Custom text"

    # Progress Bar widget
    COLOR_NO_PROGRESS = "colorNoProgress"
    SHOW_PROGRESS_TEXT = "showProgressText"
    COLOR_IN_PROGRESS = "colorInProgress"
    COLOR_COMPLETED_PROGRESS = "colorCompleted"
    COLOR_IN_PROGRESS_BACKGROUND = "colorInProgressBackground"

    # Signature widget
    SHOW_UNDO_BUTTON = "showUndoButtonValue"
    CUSTOM_UNDO_BUTTON_TEXT = "customUndoText"
    CUSTOM_TEXT = "customText"  # Used for task-navigation bar and signature widgets
    SHOW_BUTTON_VALUE = "showButtonValue"

    # Dashboard Widget
    SOURCE = "source"
    SCOPE_TAG = "scopeTag"
    SCOPE_LIST = "scopeList"
    ORDER = "order"
    SELECTED = "selected"
    SHOW_NUMBERS = "showNumbers"
    SHOW_NUMBERS_PROJECTS = "showNumbersProjects"
    SHOW_ASSIGNEES = "showAssignees"
    SHOW_ASSIGNEES_TABLE = "showAssigneesTable"
    SHOW_OPEN_TASK_ASSIGNEES = "showOpenTaskAssignees"
    SHOW_OPEN_VS_CLOSED_TASKS = "showOpenVsClosedTasks"
    SHOW_OPEN_VS_CLOSED_TASKS_ASSIGNEES = "showOpenClosedTasksAssignees"

    # Card Widget and Service Card Widget
    CUSTOM_IMAGE = "customImage"
    SHOW_IMAGE_VALUE = "showImageValue"
    CUSTOM_LINK = "customLink"
    SHOW_LINK_VALUE = "showLinkValue"

    # Task Navigation Widget
    EMPHASIZED = "emphasized"
    EMPHASIZE = "emphasize"
    LINK = "link"
    SHOW_HEIGHT_VALUE = "showHeightValue"
    TASK_BUTTONS = "taskButtons"

    # Meta Panel Widget
    SHOW_ALL = "showAll"



import json
import os

from pykechain.enums import PropertyType, Multiplicity
from pykechain.models import AttachmentProperty
from tests.classes import TestBetamax


class TestAttachment(TestBetamax):
    test_dict = {'a': 1, 'b': 3}
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/')))

    def setUp(self):
        super().setUp()
        self.property_name = 'Plot Attachment'

        self.root_model = self.project.model(name='Product')
        self.part_model = self.root_model.add_model(name='__Testing attachments', multiplicity=Multiplicity.ONE_MANY)
        self.property_model = self.part_model.add_property(
            name=self.property_name,
            property_type=PropertyType.ATTACHMENT_VALUE
        )  # type: AttachmentProperty

        self.property = self.part_model.instance().property(name=self.property_name)  # type: AttachmentProperty

    def tearDown(self):
        self.part_model.delete()
        super().tearDown()

    def test_retrieve_attachment(self):
        data = ('data.json', json.dumps(self.test_dict), 'application/json')

        self.property._upload(data)

        r = self.property.json_load()

        self.assertDictEqual(r, self.test_dict)

    def test_retrieve_value(self):
        self.assertIsNone(self.property.value)

        self.property.upload(self.project_root + '/requirements.txt')

        self.assertIsNotNone(self.property.value)

    def test_set_value_none(self):
        self.property.value = None
        self.assertFalse(self.property.has_value())

    def test_set_value_not_a_path(self):
        with self.assertRaises(FileNotFoundError):
            self.property.value = 'String must be a filepath'

    def test_upload(self):
        self.property.upload(self.project_root + '/requirements.txt')

        requirements = self.project_root + '/requirements.txt'
        self.property.upload(requirements)

    # 1.11.1
    def test_clear_an_attachment_property(self):
        # setUp
        self.property.upload(self.project_root + '/requirements.txt')

        # testing
        self.assertTrue(self.property.has_value())

        self.property.clear()
        self.assertIsNone(self.property.value)
        self.assertIsNone(self.property._value)

    def test_retrieve_filename_from_value(self):
        # setUp
        self.property.upload(self.project_root + '/requirements.txt')

        # testing
        self.assertEqual(self.property.filename, 'requirements.txt')

    def test_has_value_true(self):
        # setUp
        self.property.upload(self.project_root + '/requirements.txt')

        # testing
        self.assertTrue(self.property.has_value())

    def test_has_value_false(self):
        self.assertFalse(self.property.has_value())

    def test_add_with_properties(self):
        # setUp
        self.new_wheel = self.root_model.instance().add_with_properties(
            self.part_model,
            "new part",
            update_dict={
                self.property_name: self.project_root + '/requirements.txt'
            },
        )
